/*
Kangaroo is Copyright (C) 2001 Douglas Thain.
Permission is granted to use the source code for research and study purposes.
Other uses must be cleared with the copyright holder.
*/


#include "auth.h"
#include "debug.h"
#include "stringtools.h"

#include <stdlib.h>
#include <stdio.h>
#include <errno.h>

struct auth_ops {
	char *type;
	int (*assert)( FILE *r, FILE *w );
	int (*accept)( FILE *r, FILE *w, char **subject );
	struct auth_ops *next;
};

static struct auth_ops *list=0;

static struct auth_ops * lookup( char *type )
{
	struct auth_ops *a;

	for(a=list;a;a=a->next) {
		if(!strcmp(a->type,type)) return a;
	}

	return 0;
}

static int auth_lookup( char *subject, const char *authfile )
{
	char line[AUTH_SUBJECT_MAX];
	int length;
	FILE *f;

	f = fopen(authfile,"r");
	if(!f) return 0;

	while(fgets( line, AUTH_SUBJECT_MAX, f )) {

		if( line[0]=='#' ) continue;

		length = strlen(line);

		if( length>0 && line[length-1]=='\n' ) {
			line[length-1] = 0;
		}

		if(pattern_match(line,subject)) {
			fclose(f);
			return 1;
		}
	}

	fclose(f);
	return 0;
}

int auth_assert( FILE *r, FILE *w )
{
	char line[AUTH_LINE_MAX];
	struct auth_ops *a;
	int response;

	for( a=list; a; a=a->next ) {

		debug("auth: requesting '%s' authentication\n",a->type);

		fprintf(w,"%s\n",a->type);
		fflush(w);

		if(!fgets(line,AUTH_LINE_MAX,r)) break;
		if(sscanf(line,"%d",&response)!=1) break;

		if(response) {
			debug("auth: peer agrees to try '%s'\n",a->type);
			if(a->assert(r,w)) {
				debug("auth: successfully authenticated\n");
				if(!fgets(line,AUTH_LINE_MAX,r)) break;
				if(sscanf(line,"%d",&response)!=1) break;
				if(response) {
					debug("auth: authorized\n");
					return 1;
				} else {
					debug("auth: not authorized\n");
				}
			} else {
				debug("auth: failed to authenticate\n");
			}
		} else {
			debug("auth: peer refuses to try '%s'\n",a->type);
		}
		debug("auth: still trying...\n");
	}

	if(!a) {
		debug("auth: ran out of authenticators\n");
	} else {
		debug("auth: lost connection\n");
	}

	errno = EACCES;
	return 0;
}

int auth_accept( FILE *r, FILE *w, char **subject, const char *authfile )
{
	struct auth_ops *a;
	char type[AUTH_TYPE_MAX];

	*subject = 0;

	while(fgets(type,AUTH_TYPE_MAX,r)) {

		/* chop off the newline */
		type[strlen(type)-1]=0;

		debug("auth: peer requests '%s' authentication\n",type);

		a = lookup( type );
		if(a) {
			debug("auth: I agree to try '%s' \n",type);
			fprintf(w,"1\n");
			fflush(w);
		} else {
			debug("auth: I do not agree to '%s' \n",type);
			fprintf(w,"0\n");
			fflush(w);
			continue;
		}

		if(a->accept(r,w,subject)) {
			debug("auth: subject authenticates as %s\n",*subject);
			if(auth_lookup(*subject,authfile)) {
				debug("auth: subject is authorized\n");
				fprintf(w,"1\n");
				fflush(w);
				return 1;
			} else {
				free(*subject);
				*subject = 0;
				debug("auth: subject is not authorized\n");
				fprintf(w,"0\n");
				fflush(w);
			}
		} else {
			debug("auth: subject could not authenticate\n");
		}
		debug("auth: still trying...\n");
	}

	debug("auth: client disconnected\n");

	errno = EACCES;
	return 0;
}

int auth_register( char *type, int (*assert)(FILE*,FILE*), int (*accept)(FILE *, FILE *, char ** ) )
{
	struct auth_ops *a;

	a = (struct auth_ops *) malloc(sizeof(struct auth_ops));
	if(!a) return 0;

	a->type = type;
	a->assert = assert;
	a->accept = accept;
	a->next = list;
	list = a;

	return 1;
}

