/*
Kangaroo is Copyright (C) 2001 Douglas Thain.
Permission is granted to use the source code for research and study purposes.
Other uses must be cleared with the copyright holder.
*/


#include "config_table.h"
#include "hash_table.h"
#include "username.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#define CONFIG_TABLE_LINE_MAX 1024

struct config_table {
	struct hash_table *hash;
};

struct config_table * config_table_create() {
	struct config_table *t;

	t = malloc(sizeof(*t));
	if(!t) return 0;

	t->hash = hash_table_create(127,hash_string);
	if(!t->hash) {
		free(t);
		return 0;
	}

	return t;
}

static int delete( struct hash_table *h, char *key, void *item, void *arg )
{
	free(item);
	return 1;
}

void config_table_delete( struct config_table *t )
{
	if(t) {
		if(t->hash) {
			hash_table_iterate(t->hash,delete,0);
			hash_table_delete(t->hash);
		}
		free(t);
	}
}

int    config_table_load( struct config_table *t, const char *filename )
{
	char line[CONFIG_TABLE_LINE_MAX];
	char *key, *value;
	FILE *file;

	file = fopen(filename,"r");
	if(!file) return 0;

	while(fgets(line,CONFIG_TABLE_LINE_MAX,file)) {
		key = strtok(line," :=\t\n");
		if(!key) continue;
		if(key[0]=='#') continue;

		value = strtok(0,"\n");
		if(!value) continue;

		value = strdup(value);
		if(!value) {
			fclose(file);
			return 0;
		}

		if(!config_table_set(t,key,value)) {
			fclose(file);
			return 0;
		}
	}

	fclose(file);

	return 1;
}

int config_table_set( struct config_table *t, const char *key, const char *value )
{
	void *old;
	char *copy;

	if(value) {
		copy = strdup(value);
		if(!copy) return 0;

		if(!hash_table_insert(t->hash,key,copy,&old)) {
			free(copy);
			return 0;
		}
	} else {
		old = hash_table_remove(t->hash,key);
	}

	if(old) free(old);


	return 1;
}

static char *make_env_key( const char *key )
{
	char *ekey;
	int len;
	int i;

	ekey = strdup(key);
	if(!ekey) return 0;

	len = strlen(ekey);
	for(i=0;i<len;i++) {
		if(ekey[i]=='.') {
			ekey[i] = '_';
		} else {
			ekey[i] = toupper(ekey[i]);
		}
	}

	return ekey;
}

static const char *table_lookup( struct hash_table *h, const char *key )
{
	char *ekey;
	char *temp;

	ekey = make_env_key(key);
	if(!ekey) return 0;

	temp = getenv(ekey);
	free(ekey);
	if(temp) return temp;

	return hash_table_lookup(h,key);
}

static char * get_special( const char *key )
{
	if(!strcmp(key,"USER")) {
		char name[USERNAME_MAX];
		if(username_get(name)) {
			return strdup(name);
		} else {
			return 0;
		}
	} else {
		return 0;
	}
}

/*
First, do a simply hash lookup to get the value for a key.
Then, do macro lookups in the same table for all instances
of ${x} in the value.
*/

static char * table_lookup_recurse( struct hash_table *hash, const char *key )
{
	char *value, *subvalue, *newvalue;
	char *dollar, *lbrace, *rbrace;
	int length;

	value = get_special(key);
	if(!value) {
		value = (char*) table_lookup(hash,key);
		if(!value) return 0;

		value = strdup(value);
		if(!value) return 0;
	}

	while(1) {
		dollar = strchr(value,'$');
		if(!dollar) return value;

		lbrace = strchr(dollar,'{');
		if(!lbrace) return value;

		rbrace = strchr(lbrace,'}');
		if(!rbrace) return value;

		*rbrace = 0;

		subvalue = table_lookup_recurse(hash,lbrace+1);
		if(!subvalue) return value;

		*rbrace = '}';

		length = strlen(value) - (rbrace-dollar) + strlen(subvalue) + 1;
		newvalue = malloc(length);
		if(!newvalue) {
			free(subvalue);
			free(value);
			return 0;
		}

		*dollar = 0;

		strcpy(newvalue,value);
		strcat(newvalue,subvalue);
		strcat(newvalue,rbrace+1);

		free(subvalue);
		free(value);

		value = newvalue;
	}

	return 0;
}

char * config_table_get( struct config_table *t, const char *key )
{
	return table_lookup_recurse(t->hash,key);
}

int config_table_set_int( struct config_table *t, const char *key, int value )
{
	char str[CONFIG_TABLE_LINE_MAX];
	sprintf(str,"%d",value);
	return config_table_set(t,key,str);
}

int config_table_get_int( struct config_table *t, const char *key )
{
	char *value;
	int result;

	value = config_table_get(t,key);
	if(value) {
		result = atoi(value);
		free(value);
		return result;
	} else {
		return 0;
	}
}

static int output( struct hash_table *hash, char *key, void *item, void *arg )
{
	char *value = table_lookup_recurse(hash,key);	
	if(value) {
		fprintf(arg,"%s=%s\n",key,value);
		free(value);
	} else {
		fprintf(arg,"%s=\n",key);
	}
	return 1;
}

int config_table_save( struct config_table *t, const char *filename )
{
	FILE *file;

	file = fopen(filename,"w");
	if(!file) return 0;

	if(!hash_table_iterate(t->hash,output,file)) {
		fclose(file);
		return 0;
	}
	
	fclose(file);
	return 1;
}


