/*
 Part of multi route version of kangaroo
 Written by Gogul Balakrishnan and Rajesh Rajamani
*/ 

#include "hash_table.h"
#include "list.h"
#include "dep_graph.h"
#include "kangaroo_message.h"
#include "spool_dir.h"
#include "macros.h"
#include "reaper.h"

#include <stdio.h>
#include <errno.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

struct hash_table *dep_graph_table=0;
pthread_mutex_t hash_mutex=PTHREAD_MUTEX_INITIALIZER;


int disp_messages(struct dep_graph_node *dgn,void *arg)
{
  struct kangaroo_message *m=dgn->m;
 	debug("%d %s %s Off %d Len %d\n",m->type,m->addr,m->path,m->offset,m->length);
  return 1;
}


/*create and return a pointer to a dep_graph_node structure */
struct dep_graph_node * dep_graph_node_create() {
  
  struct dep_graph_node *dg_node;

  dg_node=(struct dep_graph_node *)malloc(sizeof(struct dep_graph_node));
  if(!dg_node)
    return 0;


  dg_node->dependents=list_create();
  if(!dg_node->dependents){
    free(dg_node);
    return 0;
  }  

  dg_node->dependent_on_count=0;
  dg_node->m=NULL;
  return dg_node;
}

int dep_graph_table_create()
{
  dep_graph_table=hash_table_create(DEP_GRAPH_TABLE_BUCKET_COUNT,hash_string);
  if(!dep_graph_table)
    return 0;
  return 1;
}


/*create and return pointer to a dep_graph structure */
struct dep_graph * dep_graph_create(char *addr,char *path) {
  struct dep_graph *dg;
  int i,j;
  
  dg=(struct dep_graph *)malloc(sizeof(struct dep_graph));
  if(!dg)
    return 0;
  
  dg->node_list=list_create();  
  if(!dg->node_list) {
    free(dg);
    return 0;
  }
  
  dg->parallel_ops=list_create();
  if(!dg->parallel_ops) {
    list_delete(dg->node_list);
    free(dg);
    return 0;
  }
  for(i=0;i<DEP_GRAPH_MAX_MOVERS;++i) {
    dg->waiting_for_push[i]=list_create();
		dg->resend_flags[i]=0;
		dg->resend_head[i]=0;
    if(!dg->waiting_for_push[i]) {
      for(j=0;j<i;++j)
	list_delete(dg->waiting_for_push[j]);
      list_delete(dg->node_list);
      list_delete(dg->parallel_ops);
      free(dg);
      return 0;
    }
  }
    
  dg->has_mover=0;
  dg->hold_dg=0;
  dg->hold=0;
  dg->has_hold_dg=0;
  dg->rename_dependent=0;
	dg->waiting_for_notification = 0;
  strcpy(dg->path,path);
  strcpy(dg->addr,addr);
  
  pthread_mutex_init(&dg->mutex,0);
	pthread_cond_init (&dg->pushcond, 0);
  return dg;
}

int dep_graph_add_dependency(void *v_dg_node_old, const void *v_dg_node_new) {

  struct dep_graph_node *dg_node_old=(struct dep_graph_node *)v_dg_node_old;
  struct dep_graph_node *dg_node_new=(struct dep_graph_node *)v_dg_node_new;
  struct kangaroo_message * oldm,*newm;
  int oldm_min,oldm_max,newm_min,newm_max;
  oldm=dg_node_old->m;
  newm=dg_node_new->m;
  switch(newm->type) {
      case KANGAROO_CREATE:
      case KANGAROO_DELETE:
      case KANGAROO_RENAME:
      case KANGAROO_TRUNCATE:
	list_push_tail(dg_node_old->dependents, dg_node_new);
	dg_node_new->dependent_on_count++;
	break;
      
      case KANGAROO_PUT:
	switch(oldm->type) {
	    case KANGAROO_CREATE:
	    case KANGAROO_DELETE:
	    case KANGAROO_RENAME:
	    case KANGAROO_TRUNCATE:
	      list_push_tail(dg_node_old->dependents, dg_node_new);
	      dg_node_new->dependent_on_count++;
	      break;
	    case KANGAROO_PUT:
	      newm_min=newm->offset;
	      newm_max=newm->offset+newm->length-1;
	      oldm_min=oldm->offset;
	      oldm_max=oldm->offset+oldm->length-1;
	      /* Verify if this condition is correct */
	      if( (newm_min>=oldm_min && newm_min<=oldm_max) ||
		  (newm_max>=oldm_min && newm_max<=oldm_max) ||
		  (newm_min<=oldm_min && newm_max>=oldm_max) ) {
		list_push_tail(dg_node_old->dependents, dg_node_new);
		dg_node_new->dependent_on_count++;

	      }
	      break;
	default:
	  return 0;
	}
	break;
      default:
	return 0;
  }
  return 1;
}
	
struct dep_graph * dep_graph_insert_message(struct kangaroo_message *m) {
  struct dep_graph *dg,*correct_dg;
  void *odg;
  struct dep_graph_node *dg_node;
  char tag[KANGAROO_LINE_MAX];
  char hash_key[KANGAROO_ADDR_MAX+KANGAROO_PATH_MAX+2];


  strcpy(hash_key,m->addr);
  strcat(hash_key,m->path);
	
  pthread_mutex_lock(&hash_mutex);
  dg=hash_table_lookup(dep_graph_table,hash_key);
  pthread_mutex_unlock(&hash_mutex);

  if(!dg){
    dg=dep_graph_create(m->addr,m->path);
    pthread_mutex_lock(&hash_mutex);
    hash_table_insert(dep_graph_table,hash_key,dg,&odg);
    pthread_mutex_unlock(&hash_mutex);
    if(!dg)
      return 0;
  }
  

  pthread_mutex_lock(&dg->mutex); 
  correct_dg=dg;
  while(correct_dg->has_hold_dg) {
    correct_dg=correct_dg->hold_dg;
  }
  
  if(!spool_dir_insert(tag,m)) {
    pthread_mutex_unlock(&dg->mutex);
    return 0;
  }
  strcpy(m->tag,tag); 
  
  if(!(dg_node=dep_graph_insert_message_internal2(correct_dg,m))) { 
    spool_dir_remove(tag); 
    pthread_mutex_unlock(&dg->mutex);
    return 0;
  }
  
  debug("Message inserted\n");
  /* What will happen when I have rename operations to hold_dg */
  
  /*
    if(correct_dg!=dg) {
    pthread_mutex_unlock(&dg->mutex);
    return dg;
  }
  */
  
  if(!dep_graph_insert_message_internal1(correct_dg,m)) {
    /*
      CAVEAT:
      Should cleanup in a different way? The problem is fixed
    */
      dep_graph_node_remove(correct_dg,dg_node);
    
      
    
    pthread_mutex_unlock(&dg->mutex);
    return 0;
  }
  pthread_mutex_unlock(&dg->mutex);
  return dg;

}

struct dep_graph * dep_graph_insert_message_without_spool(struct kangaroo_message *m) {
  struct dep_graph *dg,*correct_dg;
  void *odg;
  struct dep_graph_node *dg_node;
  char hash_key[KANGAROO_ADDR_MAX+KANGAROO_PATH_MAX+2];


  strcpy(hash_key,m->addr);
  strcat(hash_key,m->path);
	
  pthread_mutex_lock(&hash_mutex);
  dg=hash_table_lookup(dep_graph_table,hash_key);
  pthread_mutex_unlock(&hash_mutex);

  if(!dg){
    dg=dep_graph_create(m->addr,m->path);
    pthread_mutex_lock(&hash_mutex);
    hash_table_insert(dep_graph_table,hash_key,dg,&odg);
    pthread_mutex_unlock(&hash_mutex);
    if(!dg)
      return 0;
  }
  

  pthread_mutex_lock(&dg->mutex); 
  correct_dg=dg;
  while(correct_dg->has_hold_dg) {
    correct_dg=correct_dg->hold_dg;
  }
  
  
  if(!(dg_node=dep_graph_insert_message_internal2(correct_dg,m))) { 
    pthread_mutex_unlock(&dg->mutex);
    return 0;
  }
  
  debug("Message inserted\n");
  /* What will happen when I have rename operations to hold_dg */
  
  /*
    if(correct_dg!=dg) {
    pthread_mutex_unlock(&dg->mutex);
    return dg;
  }
  */
  
  if(!dep_graph_insert_message_internal1(correct_dg,m)) {
    
    /*
      CAVEAT:
      should cleanup in a different way? NO Problem is fixed 
      
    */
    dep_graph_node_remove(correct_dg,dg_node);
        
    pthread_mutex_unlock(&dg->mutex);
    return 0;
  }
  pthread_mutex_unlock(&dg->mutex);
  return dg;

}

struct dep_graph * dep_graph_insert_message_internal1(struct dep_graph *correct_dg,struct kangaroo_message *m)
 {
  char new_hash_key[KANGAROO_ADDR_MAX+KANGAROO_PATH_MAX+2];
  struct dep_graph *new_dg,*temp_dg,*correct_new_dg;
  void *odg;
  char hash_key[KANGAROO_ADDR_MAX+KANGAROO_PATH_MAX+2];

  new_dg = 0;
  strcpy(hash_key,m->addr);
  strcat(hash_key,m->path);
  
  if(m->type==KANGAROO_RENAME) {
    if(!strcmp(m->path,m->newpath)) 
      return correct_dg;
    strcpy(new_hash_key,m->addr);
    strcat(new_hash_key,m->newpath);
    
    pthread_mutex_lock(&hash_mutex);
    new_dg=hash_table_lookup(dep_graph_table,new_hash_key);
    pthread_mutex_unlock(&hash_mutex);
    
    if(!new_dg){
      /* A file of the same new name not present in dep_graph */
      new_dg=dep_graph_create(m->addr,m->newpath);
      pthread_mutex_lock(&hash_mutex);
      if(!new_dg) {
	pthread_mutex_unlock(&hash_mutex);
	return 0;
      }
      new_dg->hold=1;
      correct_dg->rename_dependent=new_dg;
      hash_table_insert(dep_graph_table,new_hash_key,new_dg,&odg);
      pthread_mutex_unlock(&hash_mutex);
      return new_dg;
    }
    
    else {
      /* A file by the same new name is found already */
      /* Should check for too many rename operations */
      /* checking if there are any delete operations in the file will help I suppose?? */
      temp_dg=dep_graph_create(m->addr,m->newpath);
      if(!temp_dg){
	return 0;
      }
      
      /* CAVEAT: No known deadlocks */
      pthread_mutex_lock(&new_dg->mutex); 
      correct_new_dg=new_dg;
      while(correct_new_dg->has_hold_dg){
	correct_new_dg=correct_new_dg->hold_dg;
      }
      
      correct_new_dg->has_hold_dg=1;
      correct_new_dg->hold_dg=temp_dg; 
      correct_new_dg->hold_dg->hold=1;
      correct_dg->rename_dependent=temp_dg;
      pthread_mutex_unlock(&new_dg->mutex);
      return temp_dg;
    }
  }
  else	{
    return correct_dg;
  }	 
 }


struct dep_graph_node * dep_graph_insert_message_internal2(struct dep_graph *dg, struct kangaroo_message *m) {

  struct dep_graph_node *dg_node;
 
  dg_node=dep_graph_node_create();
  if(!dg_node)
    return 0;
  dg_node->m=m;
  list_iterate(dg->node_list,dep_graph_add_dependency,dg_node);
  list_push_head(dg->node_list,dg_node);
  if(dg_node->dependent_on_count==0) {
    list_push_head(dg->parallel_ops,dg_node);
  }
  return dg_node;
}

int dep_graph_node_equality(void *dg1,const void *dg2) {
  return dg1==dg2;
}

int dep_graph_node_remove(struct dep_graph *dg, struct dep_graph_node *dg_node) {
  struct dep_graph_node *dg_node_dep;
  char hash_key[KANGAROO_ADDR_MAX+KANGAROO_PATH_MAX+2];
  struct dep_graph *newpath_dg,*rename_dep_dg;
  /*bug will cause deadlock 
    pthread_mutex_lock(&dg->mutex);
  */
  
  
  while((dg_node_dep=(struct dep_graph_node *)list_pop_head(dg_node->dependents))){
    if(!(--dg_node_dep->dependent_on_count)) {
      list_push_head(dg->parallel_ops,dg_node_dep);
    }
  }
  
  if(dg_node->m->type==KANGAROO_RENAME){
    if(dg->rename_dependent) {
      /* CAVEAT: Is a lock required for rename_dependent ? */
      strcpy(hash_key,dg_node->m->addr);
      strcat(hash_key,dg_node->m->newpath);
      pthread_mutex_lock(&hash_mutex);
      newpath_dg=hash_table_lookup(dep_graph_table,hash_key);
      pthread_mutex_unlock(&hash_mutex);
      
      if(newpath_dg){
	pthread_mutex_lock(&newpath_dg->mutex);
	rename_dep_dg=newpath_dg;
	while(rename_dep_dg && rename_dep_dg!=dg->rename_dependent && rename_dep_dg->has_hold_dg) {
	  rename_dep_dg=rename_dep_dg->hold_dg;
	}
	if(rename_dep_dg && rename_dep_dg==dg->rename_dependent) {
	  dg->rename_dependent->hold=0;
	  dg->rename_dependent=0;
	}
	pthread_mutex_unlock(&newpath_dg->mutex);
      }
      else
	debug("dep_graph: SUSPICIOUS OPERATION! rename_dependent cleared before rename operation\n");
    }	
    else {
      debug("dep_graph: Rename immediately removed as dependents could not be initialized\n");
    }
  }
  
  list_remove(dg->node_list,dep_graph_node_equality,dg_node);  
  list_delete(dg_node->dependents);
  spool_dir_remove(dg_node->m->tag);
  kangaroo_message_delete(dg_node->m);
  free(dg_node);
  
  /*bug: will cause deadlock  pthread_mutex_unlock(&dg->mutex);
 */ 
  return 1;
}


struct kangaroo_message * dep_graph_get_next(struct dep_graph *dg,int mover_id) {
  struct dep_graph_node *dg_node;
  struct kangaroo_message *m;

  pthread_mutex_lock(&dg->mutex);  

	if (list_size(dg->node_list) == 0)	{
		pthread_cond_broadcast(&dg->pushcond);
		pthread_mutex_unlock(&dg->mutex);
		return 0;
	}

  if(dg->hold==1) {
    pthread_mutex_unlock(&dg->mutex);
    return 0;
  }
	
  /*list_iterate(dg->parallel_ops, disp_messages,0); 
   */
  switch(dg->resend_flags[mover_id]) {
  case 0:
    dg_node=(struct dep_graph_node *)list_pop_tail(dg->parallel_ops);
    if(!dg_node) {
      pthread_mutex_unlock(&dg->mutex);
      return 0;
    }
    m=dg_node->m;
    debug("dep_graph: pushing head for mover id %d\n",mover_id);
    list_push_head(dg->waiting_for_push[mover_id],dg_node);
    break;
    
  case 1:
    dg->resend_flags[mover_id]=2;	
    dg_node=list_pop_head(dg->waiting_for_push[mover_id]);
    list_push_tail(dg->waiting_for_push[mover_id],dg_node);
    m=dg_node->m;
    break;
    
  case 2:
    if(list_peek_head(dg->waiting_for_push[mover_id])==dg->resend_head[mover_id]) {
      dg->resend_flags[mover_id]=3;
      pthread_mutex_unlock(&dg->mutex);
      return 0;
    }
    dg_node=list_pop_head(dg->waiting_for_push[mover_id]);
    list_push_tail(dg->waiting_for_push[mover_id],dg_node);
    m=dg_node->m;
    break;
    
  case 3:
    pthread_mutex_unlock(&dg->mutex);
    return 0;
    break;
  }
  pthread_mutex_unlock(&dg->mutex);
  return m;
}


/*returns a list of (struct dep_graph_node *) */
struct list *dep_graph_what_to_push(struct dep_graph *dg,int mover_id) {
  if(mover_id<DEP_GRAPH_MAX_MOVERS)
    return dg->waiting_for_push[mover_id];
  else
    return 0;
}

/* Earlier items have been pushed*/
struct dep_graph * dep_graph_pushed(struct dep_graph *dg,int mover_id){
  struct dep_graph_node *dg_node;
  struct dep_graph *child_dg;
  int i;
  
  pthread_mutex_lock(&dg->mutex);

  while((dg_node=(struct dep_graph_node *)list_pop_tail(dg->waiting_for_push[mover_id]))) {
    dep_graph_node_remove(dg,dg_node);
  }
  
  if(list_size(dg->node_list)==0 && dg->has_hold_dg==1) {
    child_dg=dg->hold_dg;
    /* CAVEAT: Do we need to lock child_dg? I dont think so */
  
    list_delete(dg->node_list);
    list_delete(dg->parallel_ops);
    
    for(i=0;i<DEP_GRAPH_MAX_MOVERS;++i)
      list_delete(dg->waiting_for_push[i]);
    
    /* copy all the elements of the child to the parent */ 
    dg->hold=child_dg->hold;
    dg->node_list=child_dg->node_list;
    dg->has_hold_dg=child_dg->has_hold_dg;
    dg->rename_dependent=child_dg->rename_dependent;
    dg->parallel_ops=child_dg->parallel_ops;
    dg->hold_dg=child_dg->hold_dg;
    
    /*address and path are going to be the same values */ 
    for(i=0;i<DEP_GRAPH_MAX_MOVERS;++i)
      dg->waiting_for_push[i]=child_dg->waiting_for_push[i];
    
    //pthread_mutex_unlock(&child_dg->mutex);//is this necessary??
    free(child_dg);
  }
  
	debug("dep_graph: received push notification\n");
	if (list_size(dg->node_list) == 0)	{
		int ret;
		debug("dep_graph: list is empty ... notify receiver\n");
		while((ret = pthread_cond_broadcast(&dg->pushcond)) != 0)	{
			debug("unable to signal waiting receiver\n");
		}
	}
  dg->resend_flags[mover_id]=0;
  dg->resend_head[mover_id]=0;
  pthread_mutex_unlock(&dg->mutex);
  return dg;
}


/* 1 is no error */
int dep_graph_rollback(struct dep_graph *dg,int mover_id, int mode){
  struct dep_graph_node *dg_node;
  pthread_mutex_lock(&dg->mutex);
  if(mode)	{
    switch(dg->resend_flags[mover_id]) {
    case 0:
      dg->resend_flags[mover_id]=1;
      dg->resend_head[mover_id]=list_peek_head(dg->waiting_for_push[mover_id]);
      break;
    case 1:
      debug("dep_graph: ROLLBACK SHOULD NOT HAPPEN IN 1\n");
      break;	
    case 2:
      dg_node=list_pop_tail(dg->waiting_for_push[mover_id]);
      list_push_head(dg->waiting_for_push[mover_id],dg_node);
      break;
    case 3:
      dg->resend_flags[mover_id]=1;
      break;
    }
  }	
  else {
    dg->resend_flags[mover_id]=0;
    while ((dg_node=(struct dep_graph_node *)list_pop_head(dg->waiting_for_push[mover_id]))) {
      list_push_tail(dg->parallel_ops,dg_node);
    }
  }

  pthread_mutex_unlock(&dg->mutex);
  return 1;
}


 
struct read_dep_args {
  struct kangaroo_message *request;
  int read_result;
  int read_done;
  }; 

static int read_iterator(void *i, const void *r )
{
	struct read_dep_args *args = (struct read_dep_args *) r;
	struct kangaroo_message *request = args->request;

	struct kangaroo_message *item = ((struct dep_graph_node *) i)->m;
	struct kangaroo_message *chunk;

	/* Wrong host?  Skip it. */
	if(strcmp(item->addr,request->addr)) {
		return 1;
	}

	switch(item->type) {
		case KANGAROO_PUT:
			/* On a matching put, continue below */
			if(!strcmp(item->path,request->path)) {
				break;
			} else {
				return 1;
			}
		case KANGAROO_CREATE:
			/* A matching create has no effect */
			return 1;
		case KANGAROO_DELETE:
			/* If a delete matches, return 'no such file' */
			if(!strcmp(item->path,request->path)) {
				args->read_result = -1;
				args->read_done = 1;
				errno = ENOENT;
				return 0;
			} else {
				return 1;
			}
		case KANGAROO_RENAME:
			/* A rename *to* the requested name causes the request to change */
			/* However, a rename *from* the requested name causes 'file not found' */

			if(!strcmp(item->newpath,request->path)) {
				strcpy(request->path,item->path);
				return 1;
			} else if(!strcmp(item->path,request->path)) {
				args->read_result = -1;
				args->read_done = 1;
				errno = ENOENT;
				return 0;				
			} else {
				return 1;
			}
		case KANGAROO_TRUNCATE:
			/* A truncate should put an upper bound on the read request */
			if(!strcmp(item->path,request->path)) {
				if((request->offset+request->length) > item->offset ) {
					request->length = MAX(0,item->offset-request->offset);
					if(request->length==0) {
						args->read_done = 1;
						args->read_result = 0;
						return 0;
					}
				}
			}
			return 1;
		default:
			return 1;
	}

	/* If we fell through, we are looking at a PUT */

	/* If the item is above the request, skip or shorten it */
	if( item->offset > request->offset ) {
		if( item->offset > (request->offset+request->length) ) {
			return 1;
		} else {
			request->length = MAX(0,item->offset - request->length);
			return 1;
		}

	/* If the item is below the request, satisfy or skip it */
	} else {
		if( (item->offset+item->length) > request->offset ) {
		  chunk = spool_dir_lookup( item->tag, 1 ); 
			if(!chunk) {
				args->read_result = -1;
				args->read_done = 1;
				return 0;
			}

			if( (item->offset+item->length) < (request->offset+request->length) ) {
				request->length = MAX(0,item->offset + item->length - request->offset);
			}

			memcpy(request->data,&chunk->data[request->offset-chunk->offset],request->length);
			kangaroo_message_delete(chunk);
			args->read_result = request->length;
			args->read_done = 1;
			return 0;
		} else {
		  return 1;
		}
	}
}



/* 0 - not able to read anything from the storage */
int dep_graph_read(struct kangaroo_message *m)
{
  struct dep_graph *dg;
  struct read_dep_args args;
  char hash_key[KANGAROO_ADDR_MAX+KANGAROO_PATH_MAX+2];
  strcpy(hash_key,m->addr);
  strcat(hash_key,m->path);

  pthread_mutex_lock(&hash_mutex);
  dg=hash_table_lookup(dep_graph_table,hash_key);
  pthread_mutex_unlock(&hash_mutex);
  
  if(!dg) {
    /* No dependency for this destination, file pair */
    return 0;
  }
  
  pthread_mutex_lock(&dg->mutex);  
  
  args.request=m;
  args.read_result=0;
  args.read_done=0;
  
  list_iterate(dg->node_list,read_iterator,&args);
  pthread_mutex_unlock(&dg->mutex);
  
	debug("dep_graph: finds in cache %d bytes\n", args.read_result);
  return args.read_result;
}


int dep_graph_has_mover (struct dep_graph *dg)	{
  return dg->has_mover;
}


void dep_graph_gets_mover (struct dep_graph *dg)	{
  pthread_mutex_lock(&dg->mutex);

  dg->has_mover = 1;

  pthread_mutex_unlock(&dg->mutex);
}


/* This should see if a hash_bucket exists with more than one messages to be sent
	 for the destination that is the same as in the message passed to it */

struct dep_graph* dep_graph_not_null(struct kangaroo_message *m)	{
  struct dep_graph *dg;
  int no_to_be_pushed=0,i;
  char hash_key[KANGAROO_ADDR_MAX+KANGAROO_PATH_MAX+2];
  strcpy(hash_key,m->addr);
  strcat(hash_key,m->path);
  
  pthread_mutex_lock(&hash_mutex);
  dg=hash_table_lookup(dep_graph_table,hash_key);
  pthread_mutex_unlock(&hash_mutex);

  if(!dg) {
    /* No dependency for this destination, file pair */
    return NULL;
  }

  pthread_mutex_lock(&dg->mutex);
 
  for(i=0;i<DEP_GRAPH_MAX_MOVERS;++i) {
    /*Contingent on the fact that I created the lists when dg was created  otherwise, check for null need to be done*/
    no_to_be_pushed+=list_size(dg->waiting_for_push[i]);
  }

  if(list_size(dg->node_list)-(list_size(dg->parallel_ops)+no_to_be_pushed)){
    pthread_mutex_unlock(&dg->mutex);
    return dg; 
  }
  else {
    pthread_mutex_unlock(&dg->mutex);
    return NULL;
  }
}



int dep_graph_number_next_hops (struct dep_graph *dg, int number) {
  if (number > DEP_GRAPH_MAX_MOVERS)
    return  DEP_GRAPH_MAX_MOVERS;
  else	
    return number;	
}



void dep_graph_delete(struct dep_graph *dg)
{
  struct dep_graph_node *dg_node;
  char hash_key[KANGAROO_ADDR_MAX+KANGAROO_PATH_MAX+2];
  int i;
  pthread_mutex_lock(&dg->mutex);

  list_delete(dg->parallel_ops);
  for(i=0;i<DEP_GRAPH_MAX_MOVERS;++i) {
    list_delete(dg->waiting_for_push[i]);
  }
	if(list_size(dg->node_list)!=0) {
		debug("dep_graph: GRAPH DELETED WHILE NOT EMPTY\n");
	}
  while((dg_node=list_pop_head(dg->node_list))){
    list_delete(dg_node->dependents);
    kangaroo_message_delete(dg_node->m);
    free(dg_node);    
  }

  if(dg->has_hold_dg){
    dep_graph_delete(dg->hold_dg);
  }
  
  list_delete(dg->node_list);
  
  strcpy(hash_key,dg->addr);
  strcat(hash_key,dg->path);
  
  pthread_mutex_lock(&hash_mutex);
  hash_table_remove(dep_graph_table,hash_key);
  pthread_mutex_unlock(&hash_mutex);
  
  pthread_mutex_unlock(&dg->mutex);
  free(dg);
  
}

int comp_msg(void *v_dg,const void *v_m) {
	struct dep_graph_node * dg_node=(struct dep_graph_node *)v_dg;
	struct kangaroo_message *m=(struct kangaroo_message *) v_m;
	if(dg_node->m==m)
		return 1;
	return 0;
}

/* wait till mover says this message has been pushed and then return */
void dep_graph_push (struct kangaroo_message *m)	{
	
  struct dep_graph *dg;
  struct dep_graph_node *dg_node;
  char hash_key[KANGAROO_ADDR_MAX+KANGAROO_PATH_MAX+2];
  strcpy(hash_key,m->addr);
  strcat(hash_key,m->path);
  dg=hash_table_lookup(dep_graph_table,hash_key);
  if(!dg) return;

	pthread_mutex_lock(&dg->mutex);
	dg->waiting_for_notification = 1;
	while (list_size(dg->node_list))	{
		pthread_cond_wait(&dg->pushcond, &dg->mutex);
	}
	dg->waiting_for_notification = 0;
	pthread_mutex_unlock(&dg->mutex);
  return;

  /*
	dg_node=(struct dep_graph_node *)list_find(dg->node_list,comp_msg,(void *)m);
  while(dg_node) {
    sleep(5);
    dg_node=(struct dep_graph_node *)list_find(dg->node_list,comp_msg,(void *)m);
  }
	while(list_size(dg->node_list)) {
		sleep(1);
	}
	*/
}

char* dep_graph_destination(struct dep_graph *dg)	{
  pthread_mutex_lock(&dg->mutex);
  /*
	dg_node=(struct dep_graph_node *)list_peek_head(dg->node_list);
  if(!dg_node){
    pthread_mutex_unlock(&dg->mutex);
    return 0;
  }
	*/
  pthread_mutex_unlock(&dg->mutex);
  return dg->addr;
}

/* return 1, if dep_graph is empty, 0 otherwise */
int dep_graph_is_empty (struct dep_graph *dg)	{
	struct dep_graph *temp_dg;
  pthread_mutex_lock(&dg->mutex);

	if (dg->waiting_for_notification)	{
		pthread_mutex_unlock(&dg->mutex);
		return 0;
	}

  if(list_size(dg->node_list)!=0){
  	pthread_mutex_unlock(&dg->mutex);
	return 0;
  }
	/* CAVEAT: Do I need to check all the descendants */
	temp_dg=dg;
	while(temp_dg->has_hold_dg) {
  	if(list_size(temp_dg->hold_dg->node_list)!=0){
    	pthread_mutex_unlock(&dg->mutex);
    	return 0;
  	}
		temp_dg=temp_dg->hold_dg;
	}

	pthread_cond_broadcast(&dg->pushcond);
  pthread_mutex_unlock(&dg->mutex);
  return 1;
}



static int loader( char *filename, char *tag, void *arg )
{
  struct kangaroo_message *m;
  
  m = spool_dir_lookup(tag,1);
  if(!m) {
    debug("loader: deleting bad chunk %s\n",filename);
    reaper_queue(filename);
    return 1;
  }
	
  if(!dep_graph_insert_message_without_spool(m)) {
    debug("loader: couldn't add %s: %s\n",filename,strerror(errno));
	  return 0;
  }
  return 1;
}

int dep_graph_load(void)
{
  return spool_dir_iterate(loader,0);
}

int hash_iterate(struct hash_table *ht, char *key, void *dg,  void *v_dg_list) {
  struct list *dg_list=(struct list *) v_dg_list;
  list_push_head(dg_list,dg);
  return 1;
}

int dep_graph_give_all_dg(struct list **p_dg_list){
  struct list *dg_list;
  dg_list=list_create();
  if(!dg_list)
    return 0;
  hash_table_iterate(dep_graph_table,hash_iterate,dg_list);
  *p_dg_list=dg_list;
  return 1;
}
































