/*
Kangaroo is Copyright (C) 2001 Douglas Thain.
Permission is granted to use the source code for research and study purposes.
Other uses must be cleared with the copyright holder.
*/


#include "error_state.h"
#include "kangaroo_types.h"
#include "kangaroo_config.h"
#include "username.h"
#include "debug.h"
#include "network.h"
#include "version.h"
#include "mover.h"

#include <pthread.h>
#include <stdarg.h>
#include <time.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>

static int  error_state=0;
static char error_string[KANGAROO_LINE_MAX];
static char error_host[KANGAROO_HOST_MAX];
static char error_file[KANGAROO_PATH_MAX];

static int    notification_sent=0;
static time_t last_notification_time=0;

static pthread_mutex_t mutex = PTHREAD_MUTEX_INITIALIZER;
static pthread_cond_t cond = PTHREAD_COND_INITIALIZER;

static time_t error_state_began=0;

void error_state_clear()
{
	pthread_mutex_lock(&mutex);
	error_state = 0;
	error_string[0] = 0;
	error_host[0] = 0;
	error_file[0] = 0;
	notification_sent = 0;
	pthread_cond_broadcast(&cond);
	pthread_mutex_unlock(&mutex);
}

int error_state_get( char *str )
{
	int result;

	pthread_mutex_lock(&mutex);

	if(error_state) {
		strcpy(str,error_string);
		result = 1;
	} else {
		result = 0;
	}

	pthread_mutex_unlock(&mutex);

	return result;
}

void error_state_set( const char *host, const char *file, const char *fmt, ... )
{
	va_list args;

	va_start(args,fmt);
	pthread_mutex_lock(&mutex);

	vsprintf(error_string,fmt,args);
	strcpy(error_host,host);
	strcpy(error_file,file);

	if(!error_state) {
		error_state = 1;
		error_state_began = time(0);
	}

	debug("error_state: %s\n",error_string);

	pthread_cond_broadcast(&cond);
	pthread_mutex_unlock(&mutex);

	va_end(args);
}

static int error_state_notify( const char *message )
{
	const char *program;
	const char *user;
	char userbuf[USERNAME_MAX];
	char command[KANGAROO_LINE_MAX];
	char host[KANGAROO_HOST_MAX];
	time_t next_notification_time;
	FILE *stream;

	program = kangaroo_config_get("kangaroo.notify.program");
	if(!program) {
		debug("error_state: kangaroo.notify.program is not set!\n");
		return 0;
	}

	user = kangaroo_config_get("kangaroo.notify.user");
	if(!user) {
		if(!username_get(userbuf)) {
			debug("error_state: couldn't determine user!\n");
			return 0;
		}
		user = userbuf;
	}

	if(!network_name_get(host)) {
		debug("error_state: couldn't determine host name!\n");
		strcpy(host,"unknown");
	}

	sprintf(command,"%s %s",program,user);

	debug("error_state: invoking %s\n",command);

	stream = popen(command,"w");
	if(!stream) return 0;

	fprintf(stream,"This is a recorded notice from the Kangaroo server at %s.\n",host);
	fprintf(stream,"I am currently unable to move data:\n");
	fprintf(stream,"\t%s\n\n",message);

	next_notification_time = time(0) + kangaroo_config_get_int("kangaroo.notify.interval");

	fprintf(stream,"I will not send you any more notices before %s\n\n",ctime(&next_notification_time));
	fprintf(stream,"%s\n",version_string());
	fprintf(stream,"%s\n",version_info());

	fclose(stream);

	notification_sent = 1;
	last_notification_time = time(0);

	return 1;
}

void * error_state_thread( void * arg )
{
	char *message;
	int delay;
	int interval;

	pthread_mutex_lock(&mutex);

	while(1) {

		delay = kangaroo_config_get_int("kangaroo.notify.delay");
		interval = kangaroo_config_get_int("kangaroo.notify.interval");

		if( error_state &&
		    !mover_is_paused() &&
                    !notification_sent &&
		    ((time(0)-last_notification_time) > interval ) && 
		    ((time(0)-error_state_began) > delay ) ) {

			message = strdup(error_string);
			if(message) {
				pthread_mutex_unlock(&mutex);
				if(error_state_notify(message)) {
					debug("error_state: sent notification to owner\n");
				} else {
					debug("error_state: couldn't send notification: %s\n",strerror(errno));
				}
				free(message);
				pthread_mutex_lock(&mutex);
			} else {
				debug("error_state: %s\n",strerror(errno));
			}

		}

		pthread_cond_wait(&cond,&mutex);
	}

	pthread_mutex_unlock(&mutex);
	return 0;
}


