/*
Kangaroo is Copyright (C) 2001 Douglas Thain.
Permission is granted to use the source code for research and study purposes.
Other uses must be cleared with the copyright holder.
*/


#include "kangaroo_client.h"
#include "kangaroo_error.h"
#include "kangaroo_config.h"
#include "kangaroo_types.h"
#include "kangaroo_message.h"

#include "network.h"
#include "network_cache.h"
#include "auth.h"

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>

struct kangaroo_server {
	FILE *rstream;
	FILE *wstream;
};

int kangaroo_name_to_address( const char *name, char *addr_string )
{
	return network_cache_name_to_addr(name,addr_string);
}

int kangaroo_address_to_name( const char *addr_string, char *name )
{
	return network_cache_addr_to_name(addr_string,name);
}

struct kangaroo_server * kangaroo_server_open( const char *host, int port )
{
	struct kangaroo_server *s=0;
	char *nearest_host=0;
	static int did_register=0;
	int net=-1;
	int save_errno;

	s = malloc(sizeof(*s));
	if(!s) return 0;

	s->rstream = s->wstream = 0;

	if(!port) {
		port = kangaroo_config_get_int("kangaroo.tcp.port");
	}

	if(!host) {
		nearest_host = kangaroo_config_get("kangaroo.nearest.host");
		host = nearest_host;
	}

	if(!host) {
		char *local_port = kangaroo_config_get("kangaroo.local.port");
		if(!local_port) goto failure;

		net = network_connect_local(local_port);
		free(local_port);

		if(net<0) {
			net = network_connect("0",port);
		}
	} else {
		net = network_connect(host,port);
	}

	if(net<0) goto failure;

	s->rstream = fdopen(net,"rb");
	if(!s->rstream) goto failure;

	s->wstream = fdopen(net,"wb");
	if(!s->wstream) goto failure;

	if(!did_register) {
		did_register = 1;
		if(!auth_trivial_register()) goto failure;
		#ifdef HAS_GLOBUS_GSS
		if(!auth_globus_register()) goto failure;
		#endif
	}

	if(!auth_assert(s->rstream,s->wstream)) goto failure;

	if(nearest_host) free(nearest_host);

	return s;

	failure:
	save_errno = errno;
	if(s) {
		if(s->rstream) fclose(s->rstream);
		if(s->wstream) fclose(s->wstream);
		if(net>=0) network_close(net);
		free(s);
	}
	if(nearest_host) free(nearest_host);
	errno = save_errno;
	return 0;
}

void kangaroo_server_close( struct kangaroo_server *s )
{
       	fclose(s->rstream);
	fclose(s->wstream);
	free(s);
}

int kangaroo_server_address( struct kangaroo_server *s, char *addr_string )
{
	network_address addr;
	int port;

	if(network_address_remote(fileno(s->rstream),&addr,&port)) {
		network_address_to_string(addr,addr_string);
		return 1;
	} else {
		return 0;
	}
}

int kangaroo_message_send( struct kangaroo_server *s, struct kangaroo_message *m )
{
	int result;

	result = kangaroo_message_write( s->wstream, m, 1 );
	if(!result) return 0;

	result = fflush(s->wstream);
	if(result!=0) return 0;

	return 1;
}

static int kangaroo_result_get( FILE *f )
{
	char line[KANGAROO_LINE_MAX];
	char *c;
	int result;
	int fields;

	c = fgets(line,KANGAROO_LINE_MAX,f);
	if(!c) return kangaroo_error_from_unix(errno);

	fields = sscanf(line,"%d",&result);
	if(fields!=1) return kangaroo_error_from_unix(errno);

	return result;
}

int kangaroo_put( struct kangaroo_server *s, const char *target_host, const char *target_path, int offset, int length, const void *data )
{
	int result;

	if(length==0) return 0;

	result = fprintf(s->wstream,"put %s %s %d %d\n",target_host,target_path,offset,length);
	if(result<0) return -1;

	result = fwrite(data,length,1,s->wstream);
	if(result!=1) return -1;
	fflush(s->wstream);

	return length;
}

int kangaroo_get( struct kangaroo_server *s, const char *target_host, const char *target_path, int offset, int length, void *data )
{
	int result;
	int actual;

	result = fprintf(s->wstream,"get %s %s %d %d\n",target_host,target_path,offset,length);
	if(result<0) return -1;
	fflush(s->wstream);

	result = kangaroo_result_get(s->rstream);
	if(result<0) {
		errno = kangaroo_error_to_unix(result);
		return -1;
	}

	if(result>0) {
		actual = fread(data,result,1,s->rstream);
		if(actual!=1) return -1;
	}

	return result;
}

int kangaroo_create( struct kangaroo_server *s, const char *target_host, const char *target_path, int mode )
{
	int result;

	result = fprintf(s->wstream,"create %s %s %d\n",target_host,target_path,mode);
	if(result<0) return -1;
	fflush(s->wstream);

	return 0;
}

int kangaroo_delete( struct kangaroo_server *s, const char *target_host, const char *target_path )
{
	int result;

	result = fprintf(s->wstream,"delete %s %s\n",target_host,target_path);
	if(result<0) return -1;
	fflush(s->wstream);

	return 0;
}

int kangaroo_truncate( struct kangaroo_server *s, const char *target_host, const char *target_path, int length )
{
	int result;

	result = fprintf(s->wstream,"truncate %s %s %d\n",target_host,target_path,length);
	if(result<0) return -1;
	fflush(s->wstream);

	return 0;
}

int kangaroo_rename( struct kangaroo_server *s, const char *target_host, const char *target_path, const char *newpath )
{
	int result;

	result = fprintf(s->wstream,"rename %s %s %s\n",target_host,target_path,newpath);
	if(result<0) return -1;
	fflush(s->wstream);

	return 0;
}

int kangaroo_commit( struct kangaroo_server *s )
{
	int result;

	result = fprintf(s->wstream,"commit\n");
	if(result<0) return -1;
	fflush(s->wstream);

	result = kangaroo_result_get(s->rstream);
	if(result<0) {
		errno = kangaroo_error_to_unix(result);
		return -1;
	} else {
		return result;
	}
}

int kangaroo_push( struct kangaroo_server *s, const char *target_host, const char *target_path )
{
	int result;

	result = fprintf(s->wstream,"push %s %s\n",target_host,target_path);
	if(result<0) return -1;
	fflush(s->wstream);

	result = kangaroo_result_get(s->rstream);
	if(result<0) {
		errno = kangaroo_error_to_unix(result);
		return -1;
	} else {
		return result;
	}
}

int kangaroo_unqueue( struct kangaroo_server *s, const char *target_host, const char *target_path )
{
	int result;

	result = fprintf(s->wstream,"unqueue %s %s\n",target_host,target_path);
	if(result<0) return -1;
	fflush(s->wstream);

	result = kangaroo_result_get(s->rstream);
	if(result<0) {
		errno = kangaroo_error_to_unix(result);
		return -1;
	} else {
		return result;
	}
}

int kangaroo_route( struct kangaroo_server *s, const char *target_host, char *hop )
{
	int result;
	int actual;

	result = fprintf(s->wstream,"route %s\n",target_host);
	if(result<0) return -1;
	fflush(s->wstream);

	result = kangaroo_result_get(s->rstream);
	if(result<0) {
		errno = kangaroo_error_to_unix(result);
		return -1;
	}

	if(result>0) {
		actual = fread(hop,result,1,s->rstream);
		if(actual!=1) {
			errno = EPROTO;
			return -1;
		}
	}

	hop[result] = 0;

	return result;
}

int kangaroo_status( struct kangaroo_server *s, char *data, int length )
{
	int result;
	int actual;

	result = fprintf(s->wstream,"status\n");
	if(result<0) return -1;
	fflush(s->wstream);

	result = kangaroo_result_get(s->rstream);
	if(result<0) {
		errno = kangaroo_error_to_unix(result);
		return -1;
	}

	if(result>0) {
		actual = fread(data,result,1,s->rstream);
		if(actual!=1) return -1;
	}

	return result;
}

int kangaroo_pause( struct kangaroo_server *s, int incoming, int outgoing )
{
	int result;

	result = fprintf(s->wstream,"pause %d %d\n",incoming,outgoing);
	if(result<0) return -1;
	fflush(s->wstream);

	result = kangaroo_result_get(s->rstream);
	if(result<0) {
		errno = kangaroo_error_to_unix(result);
		return -1;
	} else {
		return result;
	}
}

int kangaroo_resume( struct kangaroo_server *s, int incoming, int outgoing )
{
	int result;

	result = fprintf(s->wstream,"resume %d %d\n",incoming,outgoing);
	if(result<0) return -1;
	fflush(s->wstream);

	result = kangaroo_result_get(s->rstream);
	if(result<0) {
		errno = kangaroo_error_to_unix(result);
		return -1;
	} else {
		return result;
	}
}


