/*
Kangaroo is Copyright (C) 2001 Douglas Thain.
Permission is granted to use the source code for research and study purposes.
Other uses must be cleared with the copyright holder.
*/


#include "kangaroo_execute.h"
#include "kangaroo_client.h"
#include "kangaroo_error.h"
#include "kangaroo_types.h"
#include "kangaroo_summary.h"
#include "kangaroo_config.h"
#include "kangaroo_history.h"

#include "message_list.h"
#include "routing_table.h"
#include "client_cache.h"
#include "debug.h"
#include "full_io.h"
#include "macros.h"
#include "timestamp.h"
#include "reaper.h"
#include "mover.h"
#include "error_state.h"
#include "disk_info.h"

#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>

extern struct message_list *kangaroo_message_list;
extern struct routing_table *kangaroo_routing_table;
extern int sync_mode;

static int pause_incoming=0;

static void pause_if_needed()
{
	if(pause_incoming) {
		debug("incoming: paused at user request.\n");
		while(pause_incoming) {
			sleep(1);
		}
		debug("incoming: resumed at user request.\n");
	}
}

int kangaroo_execute_put_local( struct kangaroo_message *m, struct client_cache *cache )
{
	int save_errno;
	int result;
	int fd;

	switch(m->type) {
		case KANGAROO_CREATE:
			fd = open(m->path,O_CREAT|O_WRONLY,m->mode);
			if(fd<0) {
				result = fd;
				break;
			}
			close(fd);
			result = 0;
			break;
		case KANGAROO_DELETE:
			result = unlink(m->path);
			if( result<0 && errno==ENOENT ) {
				result = 0;
			}
			break;
		case KANGAROO_RENAME:
			result = rename(m->path,m->newpath);
			break;
		case KANGAROO_TRUNCATE:
			fd = open(m->path,O_CREAT|O_WRONLY);
			if(fd<0) {
				result = fd;
				break;
			}
			result = ftruncate(fd,m->offset);
			save_errno = errno;
			close(fd);
			errno = save_errno;
			break;
		case KANGAROO_PUT:
			fd = open(m->path,O_RDWR|O_CREAT,0700);
			if(fd>=0) {
				result = full_pwrite(fd,m->data,m->length,m->offset);
				save_errno = errno;
				close(fd);
				errno = save_errno;
			} else {
				result = -1;
			}
			break;
		default:
			errno = ENOSYS;
			result = -1;
			break;
	}

	return result;
}

int kangaroo_execute_put_queue( struct kangaroo_message *m, struct client_cache *cache )
{
	int delay = kangaroo_config_get_int("kangaroo.incoming.timeout.min");
	int success;
	UINT64_T avail,total;
	int reserve;

	while(1) {
		reserve = kangaroo_config_get_int("kangaroo.spool.reserve");
		success = disk_info_get(kangaroo_config_get("kangaroo.spool.dir"),&avail,&total);
		if( success==0 && (avail<reserve) ) {
			debug("incoming: available space (" UINT64_FORMAT " bytes) is less than kangaroo.spool.reserve (%d bytes)\n",avail,reserve);
			errno = ENOSPC;
		} else if(message_list_add(kangaroo_message_list,m)) {
			return m->length;
		}
		debug("incoming: couldn't queue '%s %s %s' because %s\n",kangaroo_message_name(m),m->addr,m->path,strerror(errno));
		debug("incoming: try again in %d seconds\n",delay);
		timestamp_sleep(delay*1000000);
		delay = MIN(delay*2,kangaroo_config_get_int("kangaroo.incoming.timeout.max"));
	}
	return -1;
}

int kangaroo_execute_put( struct kangaroo_message *m, struct client_cache *cache )
{
	char hop[KANGAROO_HOST_MAX];
	int result;

	pause_if_needed();

	kangaroo_history_incoming_add(m->length);

	if(message_list_contains_host(kangaroo_message_list,m->addr)) {
		return kangaroo_execute_put_queue(m,cache);
	}

	if(routing_table_get_hop(kangaroo_routing_table,m->addr,hop)) {
		if(!strcmp(hop,"@STORE")) {
			result = kangaroo_execute_put_local(m,cache);
			if(result>=0) {
				kangaroo_message_delete(m);
				return result;
			}
			debug("incoming: couldn't execute '%s %s %s' because %s\n",kangaroo_message_name(m),m->addr,m->path,strerror(errno));
			debug("incoming: queuing for later...\n");
		}
	}

	return kangaroo_execute_put_queue(m,cache);
}

static int kangaroo_execute_get_local( struct kangaroo_message *m, struct client_cache *cache ) 
{
	int fd;
	int result;
	int save_errno;

	fd = open(m->path,O_RDONLY);
	if(fd>=0) {
		result = full_pread(fd,m->data,m->length,m->offset);
		save_errno = errno;
		close(fd);
		errno = save_errno;
	} else {
		result = -1;
	}

	return result;
}

static int kangaroo_execute_get_remote( struct kangaroo_message *m, struct client_cache *cache, const char *hop ) 
{
	struct kangaroo_server *server;
	int result;
	int save_errno;

	server = client_cache_kangaroo_server_open(cache,hop,0);
	if(server) {
		result = kangaroo_get(server,m->addr,m->path,m->offset,m->length,m->data);
		if(result<0) {
			save_errno = errno;
			client_cache_kangaroo_server_close(cache);
			errno = save_errno;
		}
	} else {
		result = -1;
	}

	return result;
}

int kangaroo_execute_get( struct kangaroo_message *m, struct client_cache *cache ) 
{
	char hop[KANGAROO_HOST_MAX];
	int result;
	int delay;

	pause_if_needed();

	m->data = malloc(m->length);
	if(!m->data) return -1;

	result = message_list_read(kangaroo_message_list,m);
	if( result==0 && m->length>0 ) {
		if(routing_table_get_hop(kangaroo_routing_table,m->addr,hop)) {
			if(!strcmp(hop,"@STORE")) {
				result = kangaroo_execute_get_local(m,cache);
				if(result<0) {
					debug("get: couldn't get %s because %s\n",m->path,strerror(errno));
				}
			} else {	
				delay = kangaroo_config_get_int("kangaroo.incoming.timeout.min");
				while(1) {
					result = kangaroo_execute_get_remote(m,cache,hop);
					if(result>=0) return result;
					if(kangaroo_error_is_final(kangaroo_error_from_unix(errno))) return result;
					debug("get: couldn't get %s:%s because %s\n",m->addr,m->path,strerror(errno));
					debug("get: try again in %d seconds\n",delay);
					timestamp_sleep(delay*1000000);
					delay = MIN(delay*2,kangaroo_config_get_int("kangaroo.incoming.timeout.max"));			
				}
			}			
		} else {
			result = -1;
		}
	}

	return result;
}

static int kangaroo_execute_push_remote( struct kangaroo_message *m, struct client_cache *cache, const char *hop ) 
{
	struct kangaroo_server *server;
	int result;

	server = client_cache_kangaroo_server_open(cache,hop,0);
	if(server) {
		result = kangaroo_push(server,m->addr,m->path);
		if(result<0) {
			client_cache_kangaroo_server_close(cache);
		}
	} else {
		result = -1;
	}

	return result;
}

int kangaroo_execute_push( struct kangaroo_message *m, struct client_cache *cache )
{
	char hop[KANGAROO_HOST_MAX];
	int result;
	int delay;

	pause_if_needed();

	message_list_push(kangaroo_message_list,m);

	if(routing_table_get_hop(kangaroo_routing_table,m->addr,hop)) {
		if(!strcmp(hop,"@STORE")) {
			return 0;
		} else {
			delay = kangaroo_config_get_int("kangaroo.incoming.timeout.min");
			while(1) {
				result = kangaroo_execute_push_remote(m,cache,hop);
				if(result==0) return 0;
				errno = kangaroo_error_to_unix(result); 
				debug("push: couldn't push %s:%s because %s\n",m->addr,m->path,strerror(errno));
				debug("push: try again in %d seconds\n",delay);
				timestamp_sleep(delay*1000000);
				delay = MIN(delay*2,kangaroo_config_get_int("kangaroo.incoming.timeout.max"));			
			}
		}
	}

	return 0;
}

int kangaroo_execute_commit( struct kangaroo_message *m, struct client_cache *cache )
{
	pause_if_needed();
	sync();
	reaper_flush();
	return 0;
}

int kangaroo_execute_route( struct kangaroo_message *m, struct client_cache *cache )
{
	char hop[KANGAROO_HOST_MAX];

	if(routing_table_get_hop(kangaroo_routing_table,m->addr,hop)) {
		m->data = strdup(hop);
		if(m->data) return strlen(m->data);
	}

	return -1;
}

int kangaroo_execute_unqueue( struct kangaroo_message *m, struct client_cache *cache )
{
	errno = ENOSYS;
	return -1;
}

int kangaroo_execute_status( struct kangaroo_message *m, struct client_cache *cache )
{
	UINT64_T used, avail, total;
	UINT64_T ominute,ohour,oday, iminute,ihour,iday;

	char message[KANGAROO_LINE_MAX];
	int result;

	result = kangaroo_summary_total(kangaroo_message_list,&used);
	if(result<0) return result;

	result = disk_info_get((const char*)kangaroo_config_get("kangaroo.spool.dir"),&avail,&total);
	if(result<0) return result;

	result = error_state_get(message);
	if(!result) message[0] = 0;

	m->data = malloc(KANGAROO_LINE_MAX*4);
	if(!m->data) return -1;

	kangaroo_history_incoming_get(&iminute,&ihour,&iday);
	kangaroo_history_outgoing_get(&ominute,&ohour,&oday);

	sprintf(m->data,"[\nDiskUsed = " UINT64_FORMAT ",\nDiskAvail = " UINT64_FORMAT ",\nDiskTotal = " UINT64_FORMAT ",\nMoverError = \"%s\",\nIncomingPaused = %d,\nOutgoingPaused = %d,\nOutgoingBytesLastMinute = " UINT64_FORMAT ",\nOutgoingBytesLastHour = " UINT64_FORMAT ",\nOutgoingBytesLastDay = " UINT64_FORMAT ",\nIncomingBytesLastMinute = " UINT64_FORMAT ",\nIncomingBytesLastHour = " UINT64_FORMAT ",\nIncomingBytesLastDay = " UINT64_FORMAT "\n]\n",used,avail,total,message,pause_incoming,mover_is_paused(),ominute,ohour,oday,iminute,ihour,iday);

	return strlen(m->data);
}

int kangaroo_execute_pause( struct kangaroo_message *m, struct client_cache *cache )
{
	int incoming = atoi(m->addr);
	int outgoing = atoi(m->path);
	if(incoming) pause_incoming=1;
	if(outgoing) mover_pause();
	return 0;
}

int kangaroo_execute_resume( struct kangaroo_message *m, struct client_cache *cache )
{
	int incoming = atoi(m->addr);
	int outgoing = atoi(m->path);
	if(incoming) pause_incoming=0;
	if(outgoing) mover_resume();
	return 0;
}

