/*
Kangaroo is Copyright (C) 2001 Douglas Thain.
Permission is granted to use the source code for research and study purposes.
Other uses must be cleared with the copyright holder.
*/
/*
 Part of multi route version of kangaroo
 Modified by Gogul Balakrishnan and Rajesh Rajamani
*/ 



#include "kangaroo_execute.h"
#include "kangaroo_client.h"
#include "kangaroo_error.h"
#include "kangaroo_types.h"
#include "kangaroo_summary.h"
#include "kangaroo_config.h"
#include "kangaroo_history.h"
#include "routing_table.h"

#include "message_list.h"
#include "routing_table.h"
#include "client_cache.h"
#include "debug.h"
#include "full_io.h"
#include "macros.h"
#include "timestamp.h"
#include "reaper.h"
#include "mover.h"
#include "error_state.h"
#include "disk_info.h"
#include "dep_graph.h"

#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <sys/socket.h>
#include <pthread.h>

#define  MAX_ACTIVE_MOVERS 30

extern struct message_list *kangaroo_message_list;
extern struct routing_table *kangaroo_routing_table;
extern int sync_mode;

static int pause_incoming=0;

static int dep_graph_initialized = 0;
static int list_initialized = 0;
static int threadids_initialized = 0;
static int active_movers = 0;

static struct list *dep_graph_with_no_movers;
static pthread_t threadids[MAX_ACTIVE_MOVERS];

static pthread_mutex_t mutex = PTHREAD_MUTEX_INITIALIZER; 
pthread_mutex_t mover_receiver_lock = PTHREAD_MUTEX_INITIALIZER; 

static void pause_if_needed()
{
	if(pause_incoming) {
		debug("incoming: paused at user request.\n");
		while(pause_incoming) {
			sleep(1);
		}
		debug("incoming: resumed at user request.\n");
	}
}

/* Initialize dep_graph_table, if necessary */
int kangaroo_execute_dep_graph_initialize (void)	{

	pthread_mutex_lock(&mutex);
	if (dep_graph_initialized == 0)	{
		if ((dep_graph_initialized = dep_graph_table_create()) == 0)	{
			pthread_mutex_unlock(&mutex);
			return 0;
		}
		else dep_graph_initialized = 1;

		dep_graph_load();	//load spool files if present
		dep_graph_give_all_dg(&dep_graph_with_no_movers);
		pthread_mutex_unlock(&mutex);
		kangaroo_execute_mover_count_decrement(-1);
		return 1;
	}

	pthread_mutex_unlock(&mutex);
	return 1;
}

/* Initlize no_movers list if necessary */
static int kangaroo_execute_list_initialize (void)	{

	if ((list_initialized == 0) && (dep_graph_with_no_movers == 0))	{
		if ((dep_graph_with_no_movers =  list_create()) == 0)	{
			return 0;
		}
	}
	list_initialized = 1;
	return 1;
}

static int kangaroo_execute_threadids_initialize(void)	{
		
	if (threadids_initialized == 0)	{
		int i=0;
		threadids_initialized = 1;

		for (;i<MAX_ACTIVE_MOVERS; i++)
			threadids[i] = -1;
	}
	return 1;
		
}

static int kangaroo_execute_find_inactive_threadid(void)	{
	/* This is called with mutex lock held ... so don't lock mutex again */
	int i = 0;
	for (;i<MAX_ACTIVE_MOVERS;i++)	{
		if (threadids[i] == -1)	
			return i;
	}
		 return -1;
}

/* When there are no more packets to be moved, the mover calls this function and 
	 exits.  If this function finds that some dependency graphs have no movers, then it
	 spawns new mover_threads for those dependency graphs */
int kangaroo_execute_mover_count_decrement (int index)	{

	void *arg;
	struct dep_graph *dg;
	int rval = 0;
	int delay = kangaroo_config_get_int("kangaroo.mover.timeout.min");

	pthread_mutex_lock(&mutex);
	if (!kangaroo_execute_threadids_initialize())	{
		debug ("Error initializing threadids %s",strerror(errno));
		pthread_mutex_unlock(&mutex);
		return -1;
	}
	if (!kangaroo_execute_list_initialize())	{
		debug ("Error creating no_movers list %s",strerror(errno));
		pthread_mutex_unlock(&mutex);
		return -1;
	}

	if ((active_movers >0) && (index >= 0))	{
		threadids[index] = -1;
		rval = active_movers--;
	}
	/* If any dep_graph* present without a mover, create a new mover for that and increment
		 active_mover count -- LATER */
	dg = list_peek_head(dep_graph_with_no_movers);

	//if (dg != NULL)	{
	while ((dg != NULL)	&& (active_movers<MAX_ACTIVE_MOVERS))	{
		int tid = kangaroo_execute_find_inactive_threadid();
		if (tid == -1)	{
			debug ("inactive thread slot not found -- SOMETHING TERRIBLY WRONG\n");
			goto grave_pblm;
		}

		arg = mover_info_initialize (0,tid,dg);
		debug("starting mover for ones with no movers\n");
		while (pthread_create(&threadids[tid], 0, mover_multi_thread, arg) != 0)	{
			debug("incoming: couldn't create new thread because %s\n",strerror(errno));
			debug("incoming: try again in %d seconds\n",delay);
			timestamp_sleep(delay*1000000);
			delay = MIN(delay*2,kangaroo_config_get_int("kangaroo.incoming.timeout.max"));
		}	
		rval = active_movers++;
		dep_graph_gets_mover(dg);
		pthread_detach(threadids[tid]);
		/* Get rid of head, now that u've created a mover for it */
		list_pop_head(dep_graph_with_no_movers);	
		dg = list_peek_head(dep_graph_with_no_movers);
	}	

	grave_pblm:
	pthread_mutex_unlock(&mutex);
	return rval;
}


int kangaroo_execute_put_local( struct kangaroo_message *m, struct client_cache *cache )
{
	int save_errno;
	int result;
	int fd;

	switch(m->type) {
		case KANGAROO_CREATE:
			fd = open(m->path,O_CREAT|O_WRONLY,m->mode);
			if(fd<0) {
				result = fd;
				break;
			}
			close(fd);
			result = 0;
			break;
		case KANGAROO_DELETE:
			result = unlink(m->path);
			if( result<0 && errno==ENOENT ) {
				result = 0;
			}
			break;
		case KANGAROO_RENAME:
			result = rename(m->path,m->newpath);
			break;
		case KANGAROO_TRUNCATE:
			fd = open(m->path,O_CREAT|O_WRONLY);
			if(fd<0) {
				result = fd;
				break;
			}
			result = ftruncate(fd,m->offset);
			save_errno = errno;
			close(fd);
			errno = save_errno;
			break;
		case KANGAROO_PUT:
			fd = open(m->path,O_RDWR|O_CREAT,0700);
			if(fd>=0) {
				result = full_pwrite(fd,m->data,m->length,m->offset);
				save_errno = errno;
				close(fd);
				errno = save_errno;
			} else {
				result = -1;
			}
			break;
		default:
			errno = ENOSYS;
			result = -1;
			break;
	}

	return result;
}

int kangaroo_execute_put_queue( struct kangaroo_message *m, struct client_cache *cache )
{
	struct dep_graph *dg;
	int delay = kangaroo_config_get_int("kangaroo.incoming.timeout.min");
	int success;
	UINT64_T avail,total;
	int reserve;

	pthread_mutex_lock(&mover_receiver_lock);
	if (!kangaroo_execute_dep_graph_initialize())	{
		debug ("Error creating dep_graph_table %s",strerror(errno));
		pthread_mutex_unlock(&mover_receiver_lock);
		return -1;
	}

	while(1) {
		reserve = kangaroo_config_get_int("kangaroo.spool.reserve");
		success = disk_info_get(kangaroo_config_get("kangaroo.spool.dir"),&avail,&total);
		if( success==0 && (avail<reserve) ) {
			debug("incoming: available space (" UINT64_FORMAT " bytes) is less than kangaroo.spool.reserve (%d bytes)\n",avail,reserve);
			errno = ENOSPC;
		} 
		else if ((dg = dep_graph_insert_message(m)) != NULL)	{
			if (dep_graph_has_mover(dg))	{
				debug ("receiver: graph has mover\n");
				pthread_mutex_unlock(&mover_receiver_lock);
				return m->length;
			}
			else	{ 
				/* Find out how many active threads are present.  If less than threshold, create
					 a new thread, else put it in a list. */

				void *arg;
				int flag = 0;
				pthread_mutex_lock(&mutex);
				if (active_movers < MAX_ACTIVE_MOVERS)	{
					int tid = kangaroo_execute_find_inactive_threadid();
					if (tid == -1)	{
						debug ("inactive thread slot not found -- SOMETHING TERRIBLY WRONG\n");
						flag = 1;
						goto add_to_no_movers;
					}
					arg = mover_info_initialize(0,tid, dg);
					while (pthread_create(&threadids[tid], 0, mover_multi_thread, arg) != 0)	{
						debug("incoming: couldn't create new thread because %s\n",strerror(errno));
						debug("incoming: try again in %d seconds\n",delay);
						timestamp_sleep(delay*1000000);
						delay = MIN(delay*2,kangaroo_config_get_int("kangaroo.incoming.timeout.max"));
					}	
					active_movers++;
					dep_graph_gets_mover(dg);
					pthread_detach(threadids[tid]);
			  }	/* if < MAX_ACTIVE_MOVERS */

				//flag = 1;	

				add_to_no_movers:
				if (flag)	{ /* means u've to put it in no_movers list - list is not thread-safe!!*/
					debug("receiver: no mover for this dg %p \n", dg);
					list_push_tail(dep_graph_with_no_movers, dg);
				}	
				pthread_mutex_unlock(&mutex);
			}/* end if no_mover*/

			pthread_mutex_unlock(&mover_receiver_lock);
			return m->length;
	  }/* end insert */	

		/* else if(message_list_add(kangaroo_message_list,m)) {
			return m->length;
		} 
		*/

		debug("incoming: couldn't queue '%s %s %s' because %s\n",kangaroo_message_name(m),m->addr,m->path,strerror(errno));
		debug("incoming: try again in %d seconds\n",delay);
		timestamp_sleep(delay*1000000);
		delay = MIN(delay*2,kangaroo_config_get_int("kangaroo.incoming.timeout.max"));
	}
	return -1;
}

int kangaroo_execute_put( struct kangaroo_message *m, struct client_cache *cache )
{
	//char hop[KANGAROO_HOST_MAX];
	//int result;

	if (!kangaroo_execute_dep_graph_initialize())	{
		debug ("Error creating dep_graph_table %s",strerror(errno));
		return -1;
	}

	pause_if_needed();

	kangaroo_history_incoming_add(m->length);

	/*
	if(message_list_contains_host(kangaroo_message_list,m->addr)) {
		return kangaroo_execute_put_queue(m,cache);
	}
	if (dep_graph_not_null(m))	{
		return kangaroo_execute_put_queue(m,cache);
	}

	if(routing_table_get_hop(kangaroo_routing_table,m->addr,hop)) {
		if(!strcmp(hop,"@STORE")) {
			result = kangaroo_execute_put_local(m,cache);
			if(result>=0) {
				kangaroo_message_delete(m);
				return result;
			}
			debug("incoming: couldn't execute '%s %s %s' because %s\n",kangaroo_message_name(m),m->addr,m->path,strerror(errno));
			debug("incoming: queuing for later...\n");
		}
	}
	*/

	return kangaroo_execute_put_queue(m,cache);
}

static int kangaroo_execute_get_local( struct kangaroo_message *m, struct client_cache *cache ) 
{
	int fd;
	int result;
	int save_errno;

	fd = open(m->path,O_RDONLY);
	if(fd>=0) {
		result = full_pread(fd,m->data,m->length,m->offset);
		save_errno = errno;
		close(fd);
		errno = save_errno;
	} else {
		result = -1;
	}

	return result;
}

static int kangaroo_execute_get_remote( struct kangaroo_message *m, struct client_cache *cache, const char *hop ) 
{
	struct kangaroo_server *server;
	int result;
	int save_errno;

	server = client_cache_kangaroo_server_open(cache,hop,0);
	if(server) {
		result = kangaroo_get(server,m->addr,m->path,m->offset,m->length,m->data);
		if(result<0) {
			save_errno = errno;
			client_cache_kangaroo_server_close(cache);
			errno = save_errno;
		}
	} else {
		result = -1;
	}

	return result;
}

/* Either gets it here itself or finds it in the destination */
int kangaroo_execute_get( struct kangaroo_message *m, struct client_cache *cache ) 
{
	char hop[KANGAROO_HOST_MAX];
	int result;
	int delay;

	if (!kangaroo_execute_dep_graph_initialize())	{
		debug ("Error creating dep_graph_table %s",strerror(errno));
		return 0;
	}

	pause_if_needed();

	m->data = malloc(m->length);
	if(!m->data) return -1;

	result = dep_graph_read(m);
	debug("get: dep_graph contains %d bytes\n",result);
	if( result==0 && m->length>0 ) {
		if(routing_table_get_hop(kangaroo_routing_table,m->addr,hop)) {
			if(!strcmp(hop,"@STORE")) {
				result = kangaroo_execute_get_local(m,cache);
				if(result<0) {
					debug("get: couldn't get %s because %s\n",m->path,strerror(errno));
				}
			} else {	
				delay = kangaroo_config_get_int("kangaroo.incoming.timeout.min");
				while(1) {
					result = kangaroo_execute_get_remote(m,cache,m->addr);
					/* result = kangaroo_execute_get_remote(m,cache,hop); */
					if(result>=0) return result;
					if(kangaroo_error_is_final(kangaroo_error_from_unix(errno))) return result;
					debug("get: couldn't get %s:%s because %s\n",m->addr,m->path,strerror(errno));
					debug("get: try again in %d seconds\n",delay);
					timestamp_sleep(delay*1000000);
					delay = MIN(delay*2,kangaroo_config_get_int("kangaroo.incoming.timeout.max"));			
				}
			}			
		} else {
			result = -1;
		}
	}

	return result;
}

/* Fix this */
static int kangaroo_execute_push_remote( struct kangaroo_message *m, struct client_cache *cache, const char *hop ) 
{
	/*
	struct kangaroo_server *server;
	int result;

	server = client_cache_kangaroo_server_open(cache,hop,0);
	if(server) {
		result = kangaroo_push(server,m->addr,m->path);
		if(result<0) {
			client_cache_kangaroo_server_close(cache);
		}
	} else {
		result = -1;
	}
	*/

	int result;
	struct kangaroo_server *server;
	struct routing_table *plist;
	struct routing_table_iterator *piterator;
  char *pnxthop;

	result = 0;
	plist = routing_table_get_all_hops (kangaroo_routing_table, m->addr);
	piterator = routing_table_iterator_init(plist);

	while ((pnxthop = routing_table_iterator_nexthop(piterator)))	{
		if (pnxthop == 0)	{
			debug ("mover: nxthop is NULL!!! ... no mover initialized.\n");
			continue;
		}

		server = kangaroo_server_open(pnxthop,0);
		if (!server)	{
			result = -1;
			break;
		}
		result = kangaroo_push(server, m->addr, m->path);
		kangaroo_server_close(server);

		if (result <0)	{
			break;
		}
	}	

	return result;
}

int kangaroo_execute_push( struct kangaroo_message *m, struct client_cache *cache )
{
	char hop[KANGAROO_HOST_MAX];
	int result;
	int delay;

	pthread_mutex_lock(&mover_receiver_lock);
	if (!kangaroo_execute_dep_graph_initialize())	{
		debug ("Error creating dep_graph_table %s",strerror(errno));
		pthread_mutex_unlock(&mover_receiver_lock);
		return -1;
	}
	pthread_mutex_unlock(&mover_receiver_lock);

	pause_if_needed();
	debug ("receiver: received push for %s:%s\n", m->addr,m->path);
	dep_graph_push (m);
	debug ("receiver: sending push to next hop\n");

	/* message_list_push(kangaroo_message_list,m); */

	if(routing_table_get_hop(kangaroo_routing_table,m->addr,hop)) {
		if(!strcmp(hop,"@STORE")) {
			debug("receiver: pushing to same hop\n");
			return 0;
		} else {
			delay = kangaroo_config_get_int("kangaroo.incoming.timeout.min");
			while(1) {
				result = kangaroo_execute_push_remote(m,cache,hop);
				if(result==0) return 0;
				errno = kangaroo_error_to_unix(result); 
				debug("push: couldn't push %s:%s because %s\n",m->addr,m->path,strerror(errno));
				debug("push: try again in %d seconds\n",delay);
				timestamp_sleep(delay*1000000);
				delay = MIN(delay*2,kangaroo_config_get_int("kangaroo.incoming.timeout.max"));			
			}
		}
	}

	return 0;
}

int kangaroo_execute_commit( struct kangaroo_message *m, struct client_cache *cache )
{
	pause_if_needed();
	sync();
	reaper_flush();
	return 0;
}

int kangaroo_execute_route( struct kangaroo_message *m, struct client_cache *cache )
{
	char hop[KANGAROO_HOST_MAX];

	if(routing_table_get_hop(kangaroo_routing_table,m->addr,hop)) {
		m->data = strdup(hop);
		if(m->data) return strlen(m->data);
	}

	return -1;
}

int kangaroo_execute_unqueue( struct kangaroo_message *m, struct client_cache *cache )
{
	errno = ENOSYS;
	return -1;
}

int kangaroo_execute_status( struct kangaroo_message *m, struct client_cache *cache )
{
	UINT64_T used, avail, total;
	UINT64_T ominute,ohour,oday, iminute,ihour,iday;

	char message[KANGAROO_LINE_MAX];
	int result;

	result = kangaroo_summary_total(kangaroo_message_list,&used);
	if(result<0) return result;

	result = disk_info_get((const char*)kangaroo_config_get("kangaroo.spool.dir"),&avail,&total);
	if(result<0) return result;

	result = error_state_get(message);
	if(!result) message[0] = 0;

	m->data = malloc(KANGAROO_LINE_MAX*4);
	if(!m->data) return -1;

	kangaroo_history_incoming_get(&iminute,&ihour,&iday);
	kangaroo_history_outgoing_get(&ominute,&ohour,&oday);

	sprintf(m->data,"[\nDiskUsed = " UINT64_FORMAT ",\nDiskAvail = " UINT64_FORMAT ",\nDiskTotal = " UINT64_FORMAT ",\nMoverError = \"%s\",\nIncomingPaused = %d,\nOutgoingPaused = %d,\nOutgoingBytesLastMinute = " UINT64_FORMAT ",\nOutgoingBytesLastHour = " UINT64_FORMAT ",\nOutgoingBytesLastDay = " UINT64_FORMAT ",\nIncomingBytesLastMinute = " UINT64_FORMAT ",\nIncomingBytesLastHour = " UINT64_FORMAT ",\nIncomingBytesLastDay = " UINT64_FORMAT "\n]\n",used,avail,total,message,pause_incoming,mover_is_paused(),ominute,ohour,oday,iminute,ihour,iday);

	return strlen(m->data);
}

int kangaroo_execute_pause( struct kangaroo_message *m, struct client_cache *cache )
{
	int incoming = atoi(m->addr);
	int outgoing = atoi(m->path);
	if(incoming) pause_incoming=1;
	if(outgoing) mover_pause();
	return 0;
}

int kangaroo_execute_resume( struct kangaroo_message *m, struct client_cache *cache )
{
	int incoming = atoi(m->addr);
	int outgoing = atoi(m->path);
	if(incoming) pause_incoming=0;
	if(outgoing) mover_resume();
	return 0;
}

