/*
Kangaroo is Copyright (C) 2001 Douglas Thain.
Permission is granted to use the source code for research and study purposes.
Other uses must be cleared with the copyright holder.
*/
/*
 Part of multi route version of kangaroo
 Modified by Gogul Balakrishnan and Rajesh Rajamani
*/ 


#include "kangaroo_receive.h"
#include "kangaroo_types.h"
#include "kangaroo_execute.h"
#include "kangaroo_config.h"
#include "kangaroo_www.h"

#include "message_list.h"
#include "network.h"
#include "routing_table.h"
#include "spool_dir.h"
#include "debug.h"
#include "macros.h"
#include "version.h"
#include "auth.h"
#include "mover.h"
#include "timestamp.h"
#include "reaper.h"
#include "username.h"
#include "error_state.h"

#include <unistd.h>
#include <signal.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <pwd.h>
#include <stdlib.h>
#include <pthread.h>

#define _GNU_SOURCE
#ifdef HAS_GETOPT_H
	#include "getopt.h"
#else
	extern int getopt( int argc, char * const argv[], const char *optstring );
	extern char *optarg;
	extern int optind, opterr, optopt;
#endif

/* We load these structures once from the initial thread, and then share them among all threads */

struct message_list *kangaroo_message_list=0;
struct routing_table *kangaroo_routing_table=0;

static int local_mode=1, tcp_mode=1, www_mode=1;
static pthread_t local_tid, tcp_tid, www_tid, error_state_tid;

static void shutdown( int sig )
{
	#ifdef HAS_STRSIGNAL
	debug("shutdown: got signal %d (%s)\n",sig,strsignal(sig));
	#else
	debug("shutdown: got signal %d\n",sig);
	#endif

	if(local_mode) {
		char *local_port = kangaroo_config_get("kangaroo.local.port");
		if(local_port) {
			debug("shutdown: deleting %s...\n",local_port);
			unlink(local_port);
			free(local_port);
		}
		debug("shutdown: killing local listener thread\n");
		pthread_cancel(local_tid);
	}

	if(tcp_mode) {
		debug("shutdown: killing tcp listener thread\n");
		pthread_cancel(tcp_tid);
	}

	if(www_mode) {
		debug("shutdown: killing www listener thread\n");
		pthread_cancel(www_tid);
	}

	debug("shutdown: killing error state thread\n");
	pthread_cancel(error_state_tid);

	debug("shutdown: waiting for reaper to finish...\n");
	reaper_flush();

	debug("shutdown: exiting\n");
	exit(1);
}

static void * starter( void *arg )
{
	int fd;
	FILE *r=0, *w=0;
	char *subject=0;
	char *authfile=0;

	fd  = (int) arg;

	authfile = kangaroo_config_get("kangaroo.auth.file");
	if(!authfile) {
		fprintf(stderr,"starter: couldn't determine auth file!\n");
		goto done;
	}

	r = fdopen(fd,"rb");
	if(!r) {
		fprintf(stderr,"starter: couldn't open read stream: %s\n",strerror(errno));
		goto done;
	}

	w = fdopen(fd,"wb");
	if(!w) {
		fprintf(stderr,"starter: couldn't open write stream: %s\n",strerror(errno));
		goto done;
	}

	debug("starter: checking authorization in %s...\n",authfile);

	if(auth_accept(r,w,&subject,authfile)) {
		debug("starter: remote user %s is authorized\n",subject);
       		kangaroo_receive(r,w);
		debug("starter: %s has disconnected\n",subject);
		free(subject);
	} else {
		debug("starter: remote user is not authorized, disconnecting\n");
	}

	done:

	if(authfile) free(authfile);
	if(r) fclose(r);
	if(w) fclose(w);
	if(!r && !w) network_close(fd);

	return 0;
}

struct listener_args {
	int port;
	void * (*startfunc) ( void * );
};
	
static void *listener( void *arg )
{
	int child;
	pthread_t tid;
	struct listener_args *args = arg;

	while(1) {
		child = network_accept(args->port);
		if(child<0) {
			if(errno==EINTR) {
				continue;
			} else {
				debug("listener: couldn't accept: %s, try again in 5 seconds\n",strerror(errno));
				timestamp_sleep(5000000);
			}
		}			

		if(pthread_create( &tid, 0, args->startfunc, (void*)child )) {
			fprintf(stderr,"listener: Couldn't create thread: %s\n",strerror(errno));
		} else {
			pthread_detach( tid );
		}
	}
}

static void show_use(char *name)
{
	fprintf(stderr,"use: %s [options]\n",name);
	fprintf(stderr,"where options are:\n");
	fprintf(stderr," -c         read configuration from this file (default is %s)\n",kangaroo_config_get("kangaroo.config"));
	fprintf(stderr," -C         write configuration to this file\n");
	fprintf(stderr," -d         turn on debugging output\n");
	fprintf(stderr," -D <file>  send debug info to this file\n");
	fprintf(stderr," -f         do not fork - run in the foreground\n");
	fprintf(stderr," -l <file>  listen on this unix-domain socket (default is %s)\n",kangaroo_config_get("kangaroo.local.port"));
	fprintf(stderr," -L         do not listen on the unix-domain at all\n");
	fprintf(stderr," -s         do not sync on commit\n");
	fprintf(stderr," -t <port>  listen on this tcp port (default is %s)\n",kangaroo_config_get("kangaroo.tcp.port"));
	fprintf(stderr," -T         do not listen on tcp at all\n");
	fprintf(stderr," -u <user>  userid to run as\n");
	fprintf(stderr," -v         show version information\n");
	fprintf(stderr," -w         listen on this WWW port (default is %s)\n",kangaroo_config_get("kangaroo.www.port"));
	fprintf(stderr," -W         do not run WWW service at all\n");
}

int main( int argc, char *argv[] )
{
	int tcp_master, local_master, www_master;
	int result;
	char c;
	char *user=0;
	FILE *debug_file=0;
	char *debug_file_name=0;
	char *config_file_name=0;
	char *config_dump_file_name=0;
	int fork_mode=1;
	sigset_t set, oldset;

	struct listener_args tcp_args, local_args, www_args;

	while((c=getopt(argc,argv,"c:C:dD:fl:Lt:Tu:vw:W"))!=(char)-1) {
		switch(c) {
			case 'c':
				config_file_name = optarg;
				break;
			case 'C':
				config_dump_file_name = optarg;
				break;
			case 'd':
				debug_mode_set(1);
				break;
			case 'D':
				debug_file_name = optarg;
				break;
			case 'f':
				fork_mode = 0;
				break;
			case 'l':
				kangaroo_config_set("kangaroo.local.port",optarg);
				break;
			case 'L':
				local_mode = 0;
				break;
			case 't':
				kangaroo_config_set_int("kangaroo.tcp.port",atoi(optarg));
				break;
			case 'T':
				tcp_mode = 0;
				break;
			case 'u':
				user = optarg;
				break;
			case 'v':
				fprintf(stderr,"%s\n%s\n",argv[0],version_string());
				return 0;
			case 'w':
				kangaroo_config_set_int("kangaroo.www.port",atoi(optarg));
				break;
			case 'W':
				www_mode = 0;
				break;
			default:
				show_use(argv[0]);
				return 1;
				break;
		}
	}

	if(config_file_name) {
		result = kangaroo_config_load(config_file_name);
		if(!result) {
			fprintf(stderr,"Couldn't load configuration from %s: %s\n",config_file_name,strerror(errno));
			return 1;
		}
	}

	if(debug_file_name) {
		debug_file = fopen(debug_file_name,"w");
		if(!debug_file) {
			fprintf(stderr,"Couldn't open %s: %s\n",debug_file_name,strerror(errno));
			return 1;
		}
	} else {
		debug_file = stderr;
	}
		
	debug_stream_set(debug_file);

	if( !user && username_is_super() ) {
		fprintf(stderr,"Sorry, I refuse to run as root.\n");
		fprintf(stderr,"You must use the -u option to select a user name.\n");
		return 1;
	}
	
	if( user ) {
		if(!username_is_super() )  {
			fprintf(stderr,"You must be root to use the -u option.\n");
			return 1;
		}

		if(!username_set(user)) {
			fprintf(stderr,"couldn't change to user %s: %s\n",user,strerror(errno));
			return 1;
		}
	}

	if(!auth_trivial_register()) {
		fprintf(stderr,"Couldn't register trivial authentication: %s\n",strerror(errno));
		return 1;
	}

	#ifdef HAS_GLOBUS_GSS
	if(!auth_globus_register()) {
		fprintf(stderr,"Couldn't register Globus authentication: %s\n",strerror(errno));
		return 1;
	}
	#endif

	debug("main: loading routing table...\n");
       	kangaroo_routing_table = routing_table_load(kangaroo_config_get("kangaroo.routing.file"));
	if(!kangaroo_routing_table) {
		fprintf(stderr,"Couldn't load routing table %s: %s\n",kangaroo_config_get("kangaroo.routing.file"),strerror(errno));
		return 1;
	}

	debug("main: initializing spool dir...\n");
	result = spool_dir_init(kangaroo_config_get("kangaroo.spool.dir"));
	if(!result) {
		fprintf(stderr,"Couldn't init the spool dir %s: %s\n",kangaroo_config_get("kangaroo.spool.dir"),strerror(errno));
		return 1;
	}

	kangaroo_message_list = message_list_create();
	if(!kangaroo_message_list) {
		fprintf(stderr,"Couldn't create message list: %s\n",strerror(errno));
		return 1;
	}

	/* Should be dep_graph_load */
	/*
	
	debug("main: scanning spooled data\n");
	result = message_list_load(kangaroo_message_list);
	if(!result) {
		fprintf(stderr,"Couldn't load data from spool dir %s: %s\n",kangaroo_config_get("kangaroo.spool.dir"),strerror(errno));
		return 1;
	}
	debug("main: done\n");
	*/
	
	if(local_mode) {
		const char *local_port = kangaroo_config_get("kangaroo.local.port");
		local_master = network_serve_local( local_port );
		if(local_master<0) {
			fprintf(stderr,"Couldn't listen on %s: %s\n",local_port,strerror(errno));
			return 1;
		}
	}

	if(tcp_mode) {
		network_address addr;
		int tcp_port = kangaroo_config_get_int("kangaroo.tcp.port");
		tcp_master = network_serve( tcp_port );
		if(tcp_master<0) {
			fprintf(stderr,"Couldn't listen on tcp port %d: %s\n",tcp_port,strerror(errno));
			return 1;
		}
		/* Read back the port number in case TCP chose it for us. */
		if(!network_address_local(tcp_master,&addr,&tcp_port)) {
			fprintf(stderr,"Couldn't get address: %s\n",strerror(errno));
			return 1;
		}
		kangaroo_config_set_int("kangaroo.tcp.port",tcp_port);
	}

	if(www_mode) {
		network_address addr;
		int www_port = kangaroo_config_get_int("kangaroo.www.port");
		www_master = network_serve( www_port );
		if(www_master<0) {
			fprintf(stderr,"Couldn't listen on www port %d: %s\n",www_port,strerror(errno));
			return 1;
		}
		/* Read back the port number in case TCP chose it for us. */
		if(!network_address_local(www_master,&addr,&www_port)) {
			fprintf(stderr,"Couldn't get address: %s\n",strerror(errno));
			return 1;
		}
		kangaroo_config_set_int("kangaroo.www.port",www_port);
	}

	if(config_dump_file_name) {
		result = kangaroo_config_save(config_dump_file_name);
		if(!result) {
			fprintf(stderr,"Couldn't write configuration to %s: %s\n",config_dump_file_name,strerror(errno));
			return 1;
		}
	}

	if(fork_mode) {
		pid_t pid;
		pid = fork();
		if(pid<0) {
			fprintf(stderr,"Couldn't fork: %s\n",strerror(errno));
			return 1;
		} else if(pid>0) {
			exit(0);
		} else {
			/* child process, continue */
		}
	}

	/* By default, all spawned threads will block signals */

	sigfillset(&set);
	pthread_sigmask(SIG_BLOCK,&set,&oldset);

	if(local_mode) {
		local_args.port = local_master;
		local_args.startfunc = starter;
		debug("main: starting local port listener\n");
		pthread_create(&local_tid,0,listener,&local_args);
	}

	if(tcp_mode) {
		tcp_args.port = tcp_master;
		tcp_args.startfunc = starter;
		debug("main: starting tcp port listener\n");
		pthread_create(&tcp_tid,0,listener,&tcp_args);
	}

	if(www_mode) {
		www_args.port = www_master;
		www_args.startfunc = kangaroo_www_server;
		debug("main: starting www service\n");
		pthread_create(&www_tid,0,listener,&www_args);
	}

	debug("main: starting error state thread\n");
	pthread_create(&error_state_tid,0,error_state_thread,0);

	/* However, the master thread will accept certain signals */

	signal(SIGHUP,SIG_IGN);
	signal(SIGTERM,shutdown);
	signal(SIGINT,shutdown);
	signal(SIGQUIT,shutdown);
	signal(SIGSEGV,shutdown);

	pthread_sigmask(SIG_SETMASK,&oldset,0);

	if (!kangaroo_execute_dep_graph_initialize())	{
		debug ("Error creating dep_graph_table %s",strerror(errno));
		return -1;
	}

	while (1)	{
		sleep (5000);
	}
	exit(1);
	return 1;
}
