/*
Kangaroo is Copyright (C) 2001 Douglas Thain.
Permission is granted to use the source code for research and study purposes.
Other uses must be cleared with the copyright holder.
*/


#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <dirent.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <errno.h>

#include "kangaroo_client.h"
#include "version.h"
#include "timestamp.h"

#ifdef HAS_READLINE
#include <readline/readline.h>
#include <readline/history.h>
#endif

static UINT64_T uncommitted_bytes = 0;
static UINT64_T unpushed_bytes = 0;
static int uncommitted_ops = 0;
static int any_failure = 0;
static timestamp_t last_push_time = 0;
static timestamp_t last_commit_time = 0;

int do_get      ( struct kangaroo_server *server, char **argv );
int do_put      ( struct kangaroo_server *server, char **argv );
int do_create   ( struct kangaroo_server *server, char **argv );
int do_delete   ( struct kangaroo_server *server, char **argv );
int do_rename   ( struct kangaroo_server *server, char **argv );
int do_truncate ( struct kangaroo_server *server, char **argv );
int do_push     ( struct kangaroo_server *server, char **argv );
int do_commit   ( struct kangaroo_server *server, char **argv );
int do_route    ( struct kangaroo_server *server, char **argv );
int do_status   ( struct kangaroo_server *server, char **argv );
int do_pause    ( struct kangaroo_server *server, char **argv );
int do_resume   ( struct kangaroo_server *server, char **argv );
int do_help     ( struct kangaroo_server *server, char **argv );
int do_exit     ( struct kangaroo_server *server, char **argv );

struct command
{
	char *name;
	int (*action) ( struct kangaroo_server *server, char **argv );
	int minargs;
	int maxargs;
	int needs_commit;
	int host_arg;
	char *args;
	char *help;
};

struct command command_list[] = {
{ "get",     do_get,       3, 3, 0, 1, "<host> <file> <local-file>","retrieve a file", },
{ "put",     do_put,       3, 3, 1, 2, "<local-file> <host> <file>","send a file", },
{ "rename",  do_rename,    3, 3, 1, 1, "<host> <oldname> <newname>", "rename a remote file", },
{ "create",  do_create,    2, 3, 1, 1, "<host> <file> [mode]", "create a remote file", },
{ "delete",  do_delete,    2, 2, 1, 1, "<host> <file>", "delete a remote file", },
{ "truncate",do_truncate,  2, 3, 1, 1, "<host> <file> [size]", "truncate a remote file", },
{ "push",    do_push,      2, 2, 0, 1, "<host> <file>", "pause until all changes arrive", },
{ "commit",  do_commit,    0, 0, 0, 0, "", "pause until all changes are safe", },
{ "route",   do_route,     1, 1, 0, 1, "<host>", "display route to this host", },
{ "status",  do_status,    0, 0, 0, 0, "", "display status of this host", },
{ "pause",   do_pause,     0, 1, 0, 0, "[incoming|outgoing]", "suspend data movement", },
{ "resume",  do_resume,    0, 1, 0, 0, "[incoming|outgoing]", "continue data movement", },
{ "help",    do_help,      0, 0, 0, 0, "","show this message", },
{ "?",       do_help,      0, 0, 0, 0, "","show this message", },
{ "exit",    do_exit,      0, 0, 0, 0, "","quit the program", },
{ "quit",    do_exit,      0, 0, 0, 0, "","quit the program", },
{ 0, 0, 0, 0, 0 },
};

#define BUFFER_SIZE (1024*1024)
static char buffer[BUFFER_SIZE];

int do_get      ( struct kangaroo_server *server, char **argv )
{
	FILE *file=0;
	int actual=0;
	int lasttime = 0;
	int save_errno;
	int success;
	kangaroo_offset_t offset=0;
	timestamp_t start, stop;
	UINT64_T total=0;

	start = timestamp_get();

	if(!strcmp(argv[3],"-")) {
		file = stdout;
	} else {
		file = fopen(argv[3],"w");
		if(!file) goto failure;
	}

	while(1) {
		actual = kangaroo_get(server,argv[1],argv[2],offset,BUFFER_SIZE,buffer);
		if((actual<0) && (lasttime==0)) goto failure;		/* avoids printing error messages, if 2nd get
																											 request fails */
		if(actual==0) break;

		success = fwrite(buffer,1,actual,file);
		if(success!=actual) goto failure;

		offset+=actual;
		total+=actual;
		lasttime = 1;
	}

	if(!strcmp(argv[3],"-")) {
		/* nothing */
	} else {
		fclose(file);
	}

	stop = timestamp_get();

	printf(UINT64_FORMAT " bytes transferred in %f seconds (%f MB/s)\n",
		total,
		(stop-start)/1000000.0,
		1000000.0*total/(double)(stop-start)/1024/1024);

	return 0;

	failure:
	save_errno = errno;
	if(file) fclose(file);
	errno = save_errno;
	return -1;
}

int do_put      ( struct kangaroo_server *server, char **argv )
{
	FILE *file=0;
	int actual=0;
	int save_errno;
	int success;
	kangaroo_offset_t offset=0;
	timestamp_t start, stop;
	UINT64_T total=0;

	start = timestamp_get();

	if(!strcmp(argv[1],"-")) {
		file = stdin;
	} else {
		file = fopen(argv[1],"rb");
		if(!file) goto failure;
	}

	while(1) {
		actual = fread(buffer,1,BUFFER_SIZE,file);
		if(actual<0) goto failure;

		if(actual==0) break;

		success = kangaroo_put(server,argv[2],argv[3],offset,actual,buffer);
		if(success!=actual) goto failure;

		offset += actual;

		uncommitted_bytes += actual;
		unpushed_bytes += actual;
		total += actual;
	}

	if(!strcmp(argv[1],"-")) {
		/* nothing */
	} else {
		fclose(file);
	}

	stop = timestamp_get();

	printf(UINT64_FORMAT " bytes inserted in %f seconds (%f MB/s)\n",
		total,
		(stop-start)/1000000.0,
		1000000.0*total/(double)(stop-start)/1024/1024);

	return 0;

	failure:
	save_errno = errno;
	if(file) fclose(file);
	errno = save_errno;
	return -1;
}

int do_create   ( struct kangaroo_server *server, char **argv )
{
	int mode;

	if(argv[3]) {
		mode = atoi(argv[3]);
	} else {
		mode = 0777;
	}

	return kangaroo_create(server,argv[1],argv[2],mode);
}

int do_delete   ( struct kangaroo_server *server, char **argv )
{
	return kangaroo_delete(server,argv[1],argv[2]);
}

int do_rename   ( struct kangaroo_server *server, char **argv )
{
	return kangaroo_rename(server,argv[1],argv[2],argv[3]);
}

int do_truncate ( struct kangaroo_server *server, char **argv )
{
	int size;

	if(argv[3]) {
		size = atol(argv[3]);
	} else {
		size = 0;
	}

	return kangaroo_truncate(server,argv[1],argv[2],size);
}

int do_push     ( struct kangaroo_server *server, char **argv )
{
	int result;
	result = kangaroo_push(server,argv[1],argv[2]);
	if(result>=0) {
		timestamp_t current = timestamp_get();
		timestamp_t elapsed = current - last_push_time;

		printf(UINT64_FORMAT " bytes pushed in %f seconds (%f MB/s)\n",
			unpushed_bytes,
			elapsed/1000000.0,
			1000000.0*unpushed_bytes/(double)(elapsed)/1024/1024);

		last_push_time = last_commit_time = current;
		uncommitted_bytes = 0;
		uncommitted_ops = 0;
		unpushed_bytes = 0;
	}
	return result;
}

int do_commit   ( struct kangaroo_server *server, char **argv )
{
	int result;
	result = kangaroo_commit(server);
	if(result==0) {
		timestamp_t current = timestamp_get();
		timestamp_t elapsed = current - last_commit_time;

		printf(UINT64_FORMAT " bytes committed in %f seconds (%f MB/s)\n",
			uncommitted_bytes,
			elapsed/1000000.0,
			1000000.0*uncommitted_bytes/(double)(elapsed)/1024/1024);

		last_commit_time = current;
		uncommitted_bytes = 0;
		uncommitted_ops = 0;
	}
	return result;
}

int do_route    ( struct kangaroo_server *server, char **argv )
{
	char hop[KANGAROO_HOST_MAX];
	char name[KANGAROO_HOST_MAX];
	struct kangaroo_server *s;
	int result;
	int save_errno;

	char *n;

	if(!kangaroo_server_address(server,hop)) return -1;
	if(kangaroo_address_to_name(hop,name)) {
		n = name;
	} else {
		n = "unknown";
	}

	printf("nearest is %s (%s)\n",hop,n);

	while( 1 ) {
		s = kangaroo_server_open(hop,0);
		if(!s) return -1;
		result = kangaroo_route(s,argv[1],hop);
		if(result<0) {
			save_errno = errno;
			kangaroo_server_close(s);
			errno = save_errno;
			return -1;
		}

		if(hop[0]=='@') break;

		if(kangaroo_address_to_name(hop,name)) {
			n = name;
		} else {
			n = "unknown";
		}

		printf("next is %s (%s)\n",hop,n);
	}

	return 0;
}

int do_status ( struct kangaroo_server *server, char **argv )
{
	char buffer[BUFFER_SIZE];
	int result;

	result = kangaroo_status(server,buffer,BUFFER_SIZE);
	if(result>0) {
		buffer[result] = 0;
		printf(buffer);
	}
	return result;
}

int do_pause   ( struct kangaroo_server *server, char **argv )
{
	int incoming=0, outgoing=0;

	if(!argv[1]) {
		incoming = outgoing = 1;
	} else if(!strcmp(argv[1],"incoming")) {
		incoming = 1;
	} else if(!strcmp(argv[1],"outgoing")) {
		outgoing = 1;
	} else {
		errno = EINVAL;
		return -1;
	}

	return kangaroo_pause(server,incoming,outgoing);
}

int do_resume   ( struct kangaroo_server *server, char **argv )
{
	int incoming=0, outgoing=0;

	if(!argv[1]) {
		incoming = outgoing = 1;
	} else if(!strcmp(argv[1],"incoming")) {
		incoming = 1;
	} else if(!strcmp(argv[1],"outgoing")) {
		outgoing = 1;
	} else {
		errno = EINVAL;
		return -1;
	}

	return kangaroo_resume(server,incoming,outgoing);
}

int do_help     ( struct kangaroo_server *server, char **argv )
{
	struct command *c;

	for( c=command_list; c->name; c++ ) {
		printf("%10s %-30s %s\n",c->name,c->args,c->help);
	}
	printf("\n%s\n%s\n",version_string(),version_info());

	return 0;
}

int do_exit     ( struct kangaroo_server *server, char **argv )
{
	if(uncommitted_ops) {
		fprintf(stderr,"Kangaroo Warning: You didn't 'commit' before exiting.\n");
		fprintf(stderr,"Kangaroo Warning: I can't guarantee your data will be safely delivered.\n");
	}

	if(any_failure) {
		fprintf(stderr,"Kangaroo Warning: %d commands failed in this session\n",any_failure);
		exit(1);
	} else {
		exit(0);
	}

	return 0;
}

#ifndef HAS_READLINE

char * readline( char *prompt )
{
	char line[KANGAROO_LINE_MAX];

	printf("%s",prompt);
	fflush(stdout);

	if(fgets(line,KANGAROO_LINE_MAX,stdin)) {
		return strdup(line);
	} else {
		return 0;
	}
}

void add_history( const char *line )
{
}

#endif

static char ** parseline( char *line, int *argc )
{
	int maxargs=10;
	char **argv;

	*argc=0;

	argv = (char**) malloc(maxargs*sizeof(char*));
	if(!argv) return 0;

	for( (*argc)=0; (*argc)<maxargs; (*argc)++ ) {
		argv[*argc] = strtok(line," \t");
		if(line) line=0;
		if(!argv[*argc]) break;
	}

	if( (*argc)==maxargs ) {
		argv[(*argc)-1] = 0;
	}

	return argv;
}

static void run_command( struct kangaroo_server *server, struct command *c, int argc, char **argv )
{
	char addr[KANGAROO_ADDR_MAX];
	int reply;

	if( argc<c->minargs || argc>c->maxargs ) {
		printf("%s: wrong number of arguments (type 'help' for help)\n",argv[0]);
		any_failure++;
		return;
	}

	if( c->host_arg>0 ) {
		if(kangaroo_name_to_address(argv[c->host_arg],addr)) {
			argv[c->host_arg] = addr;
		} else {
			any_failure++;
			printf("couldn't find host name %s: %s\n",argv[c->host_arg],strerror(errno));
			return;

		}
	}

	reply = c->action(server,argv);
	if(reply<0) {
		printf("%s: %s\n",argv[0],strerror(errno));
		any_failure++;
	} else {
		if(c->needs_commit) {
			uncommitted_ops++;
		}
	}
}

int main( int argc, char *argv[] )
{
	char *host=0;
	struct kangaroo_server *server;
	char *line=0;
	char **arg_list=0;
	int arg_count;
	struct command *c;
	char addr[KANGAROO_ADDR_MAX];

	if(argc>2) {
		fprintf(stderr,"use: %s [nearest-host]\n",argv[0]);
		return 1;
	}

	if(argc==2) {
		host = argv[1];
	}

	server = kangaroo_server_open(host,0);
	if(!server) {
		fprintf(stderr,"couldn't contact %s: %s\n", host ? host : "nearest server", strerror(errno));
		return 1;
	}

	if(kangaroo_server_address(server,addr)) {
		printf("connected to %s\n",addr);
	}

	last_push_time = last_commit_time = timestamp_get();

	while(1) {
		if(line) {
			free(line);
			line = 0;
		}

		if(arg_list) {
			free(arg_list);
			arg_list = 0;
		}

	        line = readline( " kangaroo> " );
		if(!line) break;

		add_history(line);

		arg_list = parseline(line,&arg_count);
		if(!arg_list) continue;
		if(!arg_list[0]) continue;
		if(!*arg_list[0]) continue;

		arg_count--;

		for( c=command_list; c->name; c++ ) {
			if(!strcmp(arg_list[0],c->name)) {
				run_command(server,c,arg_count,arg_list);
				break;
			}
		}

		if( !c->name ) {
			printf("%s: unknown command (type 'help' for help)\n",arg_list[0]);
		}
	}

	printf("\n");
	do_exit(server,0);
	return 0;
}
