/*
Kangaroo is Copyright (C) 2001 Douglas Thain.
Permission is granted to use the source code for research and study purposes.
Other uses must be cleared with the copyright holder.
*/


#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <stdlib.h>

#include "kangaroo_www.h"
#include "kangaroo_summary.h"
#include "kangaroo_error.h"

#include "message_list.h"
#include "network.h"
#include "username.h"
#include "version.h"
#include "debug.h"
#include "stringtools.h"
#include "error_state.h"

extern struct message_list *kangaroo_message_list;

static void output_mover_line( FILE *stream )
{
	char message[KANGAROO_LINE_MAX];
	if(error_state_get(message)) {
		fprintf(stream,"<font color=red><b>Kangaroo is stopped: %s</b></font>",message);
	} else {
		fprintf(stream,"Kangaroo is running normally.\n");
	}
}

static int output_message_line( struct kangaroo_message *m, void *arg )
{
	FILE *stream = (FILE*) arg;
	char bytes[KANGAROO_LINE_MAX];

	if(m->length) {
		metric_units(m->length,-1,bytes);
		strcat(bytes,"B");
	} else {
		strcpy(bytes,"0");
	}

	fprintf(stream,"<tr><td>%s<td>%s<td>%s<td>1<td>%s\n",
		m->addr,m->path,kangaroo_message_name(m),bytes);

	return 1;
}


static int output_summary_line( FILE *stream, struct kangaroo_summary *s )
{
	char bytes[KANGAROO_LINE_MAX];

	if(s->bytes) {
		metric_units(s->bytes,-1,bytes);
		strcat(bytes,"B");
	} else {
		strcpy(bytes,"0");
	}

	fprintf(stream,"<tr><td>%s<td>%s<td>*<td>%ld<td>%s\n",
		s->host,s->path,s->ops,bytes);

	return 1;
}

static void handle_request( const char *path, FILE *stream )
{
	char bytes[KANGAROO_LINE_MAX];
	char host[NETWORK_NAME_MAX];
	char user[USERNAME_MAX];

	enum kangaroo_summary_type type;
	struct kangaroo_summary *s;
	struct list *l;

	long total_messages=0;
	long total_bytes=0;

	if(!network_name_get(host)) {
		strcpy(host,"unknown");
	}

	if(!username_get(user)) {
		strcpy(user,"unknown");
	}

	fprintf(stream,"HTTP/1.0 200 OK\n");
	fprintf(stream,"Server: %s\n",version_string());
	fprintf(stream,"Connection: close\n");
	fprintf(stream,"Expires: 0\n");
	fprintf(stream,"Content-Type: text/html\n\n");
	fprintf(stream,"<head><title>%s kangaroo server</title></head>\n",host);
	fprintf(stream,"<body bgcolor=white>\n");
	fprintf(stream,"<h1>%s kangaroo server</h1>\n",host);
	fprintf(stream,"%s ",version_string());
	fprintf(stream,"<a href=http://www.cs.wisc.edu/condor/kangaroo>Web Page</a>, ");
	fprintf(stream,"<a href=http://www.cs.wisc.edu/condor/kangaroo/kangaroo-%d_%d/doc/kangaroo.html>Manual</a>, ",version_major(),version_minor());
	fprintf(stream,"<a href=mailto:%s@%s>Operator</a><br>\n",user,host);
	fprintf(stream,"<hr>\n");
	output_mover_line(stream);
	fprintf(stream,"<hr>\n");
	fprintf(stream,"View by: <a href=/host>Host</a> <a href=/file>File</a> <a href=/message>Message</a>\n");
	fprintf(stream,"<hr>\n");
	fprintf(stream,"<h2>Queued Messages:</h2>\n");

	fprintf(stream,"<table border=1>\n");
	fprintf(stream,"<tr><td><b>Destination</b><td><b>File</b><td><b>Message</b><td><b>Number</b><td><b>Size</b>\n");

	if( !strcasecmp(path,"/message") ) {
		message_list_iterate_reverse( kangaroo_message_list, output_message_line, stream );
	} else {
		if( !strcasecmp(path,"/file") ) {
			type = KANGAROO_SUMMARY_BY_FILE;
		} else {
			type = KANGAROO_SUMMARY_BY_HOST;
		}

		l = kangaroo_summary_create(kangaroo_message_list,type);
		if(!l) {
			fprintf(stream,"<b>Couldn't generate summary: %s</b>",strerror(errno));
		} else {
			while( s = list_pop_head(l) ) {
				output_summary_line(stream,s);
				total_messages += s->ops;
				total_bytes += s->bytes;
				free(s);
			}
			list_delete(l);
		}
	}

	if(total_bytes) {
		metric_units(total_bytes,-1,bytes);
		strcat(bytes,"B");
	} else {
		strcpy(bytes,"0");
	}

	fprintf(stream,"<tr><td><b>Total:</b><td><td><td>%ld<td>%s\n",total_messages,bytes);
	fprintf(stream,"</table>\n");
	fprintf(stream,"</body>\n");
}

void * kangaroo_www_server( void *arg )
{
	int fd;
	int fields;
	FILE *stream;
	char action[KANGAROO_LINE_MAX];
	char path[KANGAROO_LINE_MAX];
	char version[KANGAROO_LINE_MAX];
	char line[KANGAROO_LINE_MAX];

	fd = (int) arg;

	stream = fdopen(fd,"r+");
	if(!stream) {
		close(fd);
		return 0;
	}

	if(!fgets(line,KANGAROO_LINE_MAX,stream)) {
		debug("www: empty request\n");
		goto done;
	}

	fields = sscanf(line,"%s %s %s",action,path,version);
	if(fields!=3) {
		debug("www: invalid request: %s\n",line);
		goto done;
	}

	if(strcasecmp(action,"get")) {
		debug("www: can't handle request of type %s\n",action);
		goto done;
	}
	
	while(fgets(line,KANGAROO_LINE_MAX,stream)) {
		chomp(line);
		if(!line[0]) break;
	}

	handle_request(path,stream);

	done:
	fclose(stream);
	return 0;
}

