/*
Kangaroo is Copyright (C) 2001 Douglas Thain.
Permission is granted to use the source code for research and study purposes.
Other uses must be cleared with the copyright holder.
*/


#include <pthread.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <unistd.h>

#include "kangaroo_message.h"

#include "message_list.h"
#include "spool_dir.h"
#include "list.h"
#include "macros.h"
#include "debug.h"
#include "reaper.h"
#include "hash_table.h"

struct message_list {
	struct list *incoming;
	struct list *outgoing;
	struct hash_table *table;
	pthread_mutex_t mutex;
	pthread_cond_t cond;
	int read_result;
	int read_done;
};

struct host_entry {
	int ops;
	int bytes;
};

struct message_list * message_list_create()
{
	struct message_list *l;

	l = malloc(sizeof(*l));
	if(!l) return 0;

	l->incoming = list_create();
	if(!l->incoming) {
		free(l);
		return 0;
	}

	l->outgoing = list_create();
	if(!l->outgoing) {
		list_delete(l->incoming);
		free(l);
		return 0;
	}

	l->table = hash_table_create(127,hash_string);
	if(!l->table) {
		list_delete(l->outgoing);
		list_delete(l->incoming);
		free(l);
		return 0;
	}

	pthread_mutex_init(&l->mutex,0);
	pthread_cond_init(&l->cond,0);

	return l;
}

static int message_list_add_internal( struct message_list *l, struct kangaroo_message *m )
{
	struct host_entry *h;
	void *old;

	h = hash_table_lookup(l->table,m->addr);
	if(!h) {
		h = malloc(sizeof(*h));
		if(!h) return 0;
		memset(h,0,sizeof(*h));
		if(!hash_table_insert(l->table,m->addr,h,&old)) {
			free(h);
			return 0;
		}
	}

	if(!list_push_head(l->incoming,m)) return 0;

	h->ops++;
	h->bytes += m->length;

	kangaroo_message_release_data(m);

	return 1;
}

static int loader( char *filename, char *tag, void *arg )
{
        struct kangaroo_message *m;
	struct message_list *l = (struct message_list *) arg;
	
        m = spool_dir_lookup(tag,0);
        if(!m) {
                debug("loader: deleting bad chunk %s\n",filename);
                reaper_queue(filename);
                return 1;
        }

        if(!message_list_add_internal(l,m)) {
                debug("loader: couldn't add %s: %s\n",filename,strerror(errno));
                return 0;
        }
        return 1;
}

int message_list_load( struct message_list *l )
{
	return spool_dir_iterate(loader,l);
}

int message_list_add( struct message_list *l, struct kangaroo_message *m )
{
	char tag[KANGAROO_LINE_MAX];

	pthread_mutex_lock(&l->mutex);

	if(!spool_dir_insert(tag,m)) {
		pthread_mutex_unlock(&l->mutex);
		return 0;
	}

	strcpy(m->tag,tag);
	if(!message_list_add_internal(l,m)) {
		spool_dir_remove(tag);
		pthread_mutex_unlock(&l->mutex);
		return 0;
	}

	pthread_cond_broadcast(&l->cond);	
	pthread_mutex_unlock(&l->mutex);

	return 1;
}

struct read_args {
	struct kangaroo_message *request;
	struct message_list *list;
};

int read_iterator( void *i, const void *r )
{
	struct read_args *args = (struct read_args *) r;
	struct kangaroo_message *request = args->request;
	struct message_list *list = args->list;
	struct kangaroo_message *item = (struct kangaroo_message *) i;
	struct kangaroo_message *chunk;

	/* Wrong host?  Skip it. */
	if(strcmp(item->addr,request->addr)) {
		return 1;
	}

	switch(item->type) {
		case KANGAROO_PUT:
			/* On a matching put, continue below */
			if(!strcmp(item->path,request->path)) {
				break;
			} else {
				return 1;
			}
		case KANGAROO_CREATE:
			/* A matching create has no effect */
			return 1;
		case KANGAROO_DELETE:
			/* If a delete matches, return 'no such file' */
			if(!strcmp(item->path,request->path)) {
				list->read_result = -1;
				list->read_done = 1;
				errno = ENOENT;
				return 0;
			} else {
				return 1;
			}
		case KANGAROO_RENAME:
			/* A rename *to* the requested name causes the request to change */
			/* However, a rename *from* the requested name causes 'file not found' */

			if(!strcmp(item->newpath,request->path)) {
				strcpy(request->path,item->path);
				return 1;
			} else if(!strcmp(item->path,request->path)) {
				list->read_result = -1;
				list->read_done = 1;
				errno = ENOENT;
				return 0;				
			} else {
				return 1;
			}
		case KANGAROO_TRUNCATE:
			/* A truncate should put an upper bound on the read request */
			if(!strcmp(item->path,request->path)) {
				if((request->offset+request->length) > item->length ) {
					request->length = MAX(0,item->length-request->offset);
					if(request->length==0) {
						list->read_done = 1;
						list->read_result = 0;
						return 0;
					}
				}
			}
			return 1;
		default:
			return 1;
	}

	/* If we fell through, we are looking at a PUT */

	/* If the item is above the request, skip or shorten it */
	if( item->offset > request->offset ) {
		if( item->offset > (request->offset+request->length) ) {
			return 1;
		} else {
			request->length = MAX(0,item->offset - request->length);
			return 1;
		}

	/* If the item is below the request, satisfy or skip it */
	} else {
		if( (item->offset+item->length) > request->offset ) {
			chunk = spool_dir_lookup( item->tag, 1 );
			if(!chunk) {
				list->read_result = -1;
				list->read_done = 1;
				return 0;
			}

			if( (item->offset+item->length) < (request->offset+request->length) ) {
				request->length = MAX(0,item->offset + item->length - request->offset);
			}

			memcpy(request->data,&chunk->data[request->offset-chunk->offset],request->length);
			kangaroo_message_delete(chunk);
			list->read_result = request->length;
			list->read_done = 1;
			return 0;
		} else {
			return 1;
		}
	}
}

int message_list_contains_host( struct message_list *l, const char *addr )
{
	return hash_table_lookup(l->table,addr)!=0;	
}

int message_list_read( struct message_list *l, struct kangaroo_message *m )
{
	struct read_args args;
	pthread_mutex_lock(&l->mutex);
	args.list = l;
	args.request = m;
	l->read_done = 0;
	l->read_result = 0;
	list_iterate(l->incoming,read_iterator,&args);
	if(!l->read_done) list_iterate(l->outgoing,read_iterator,&args);
	pthread_mutex_unlock(&l->mutex);
	return l->read_result;
}

void message_list_push( struct message_list *l, struct kangaroo_message *m )
{
	pthread_mutex_lock(&l->mutex);
	while(list_size(l->incoming) || list_size(l->outgoing)) {
		pthread_cond_wait(&l->cond,&l->mutex);
	}
	pthread_mutex_unlock(&l->mutex);
}

struct kangaroo_message * message_list_consume( struct message_list *l )
{
	struct kangaroo_message *m;
	pthread_mutex_lock(&l->mutex);
	while(	!(m = list_pop_tail(l->incoming)) ) {
		pthread_cond_wait(&l->cond,&l->mutex);
	}
	list_push_head(l->outgoing,m);
	pthread_mutex_unlock(&l->mutex);
	return m;
}

int message_list_consume_peek( struct message_list *l )
{
	struct kangaroo_message *m;
	pthread_mutex_lock(&l->mutex);
	m = list_peek_tail(l->incoming);
	pthread_mutex_unlock(&l->mutex);
	return m!=0;
}

int message_list_consume_commit( struct message_list *l )
{
	struct kangaroo_message *m;
	struct host_entry *h;

	pthread_mutex_lock(&l->mutex);

	while( (m=list_pop_tail(l->outgoing)) ) {
		spool_dir_remove(m->tag);
		h = hash_table_lookup(l->table,m->addr);
		if(h) {
			h->ops--;
			h->bytes -= m->length;
			if(!h->ops) {
				hash_table_remove(l->table,m->addr);
				free(h);
			}
		}
		kangaroo_message_delete(m);
	}

	pthread_cond_broadcast(&l->cond);
	pthread_mutex_unlock(&l->mutex);

	return 1;
}

void message_list_consume_abort( struct message_list *l )
{
	struct kangaroo_message *m;
	pthread_mutex_lock(&l->mutex);
	while( (m=list_pop_head(l->outgoing)) ) {
		list_push_tail(l->incoming,m);
	}
	pthread_mutex_unlock(&l->mutex);
}

int message_list_iterate( struct message_list *l, message_list_op_t iterator, void *arg )
{
	int keepgoing;
	pthread_mutex_lock(&l->mutex);
	keepgoing = list_iterate(l->incoming,(list_op_t)iterator,arg);
	if(keepgoing) {
		keepgoing = list_iterate(l->outgoing,(list_op_t)iterator,arg);
	}
	pthread_mutex_unlock(&l->mutex);
	return keepgoing;
}

int message_list_iterate_reverse( struct message_list *l, message_list_op_t iterator, void *arg )
{
	int keepgoing;
	pthread_mutex_lock(&l->mutex);
	keepgoing = list_iterate_reverse(l->outgoing,(list_op_t)iterator,arg);
	if(keepgoing) {
		keepgoing = list_iterate_reverse(l->incoming,(list_op_t)iterator,arg);
	}
	pthread_mutex_unlock(&l->mutex);
	return keepgoing;
}

