/*
Kangaroo is Copyright (C) 2001 Douglas Thain.
Permission is granted to use the source code for research and study purposes.
Other uses must be cleared with the copyright holder.
*/


#include <pthread.h>
#include <time.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <unistd.h>

#include "kangaroo_types.h"
#include "kangaroo_config.h"
#include "kangaroo_client.h"
#include "kangaroo_execute.h"
#include "kangaroo_history.h"

#include "message_list.h"
#include "spool_dir.h"
#include "routing_table.h"
#include "debug.h"
#include "mover.h"
#include "macros.h"
#include "timestamp.h"
#include "error_state.h"

static struct kangaroo_server *server=0;
static char last_hop[KANGAROO_HOST_MAX] = { 0 };
static time_t last_commit = 0;
static int messages_out = 0;
static int pause_mode = 0;

extern struct routing_table *kangaroo_routing_table;
extern struct message_list *kangaroo_message_list;

void mover_pause()
{
	pause_mode = 1;
	error_state_clear();
}

void mover_resume()
{
	pause_mode = 0;
	error_state_clear();
}

int mover_is_paused()
{
	return pause_mode;
}

static void mover_abort()
{
	message_list_consume_abort(kangaroo_message_list);
	if(server) {
		kangaroo_server_close(server);
		server = 0;
	}
	messages_out = 0;
}

static void mover_commit_without_delete()
{
	messages_out = 0;
	last_commit = time(0);
	error_state_clear();
}

static void mover_commit()
{
	kangaroo_execute_commit(0,0);
	message_list_consume_commit(kangaroo_message_list);
	messages_out = 0;
	last_commit = time(0);
	error_state_clear();
}

static int mover_send( struct kangaroo_message *incore )
{
	struct kangaroo_message *m;
	char hop[KANGAROO_HOST_MAX];
	int save_errno;
	int result;

	m = spool_dir_lookup( incore->tag, 1 );
	if(!m) return 0;


	result = routing_table_get_hop( kangaroo_routing_table, m->addr, hop );
	if(!result) {
		strcpy(hop,"unknown");
		goto failure;
	}

	if(hop[0]=='@') {
		result = kangaroo_execute_put_local(m,0);
		if(result>=0) {
			goto success;
		} else {
			goto failure;
		}
	}

	if(server && strcmp(hop,last_hop)) {
		debug("mover: changing from %s to %s, so committing.\n",hop,last_hop);
		if(kangaroo_commit(server)==0) {
			mover_commit();
			kangaroo_server_close(server);
			server = 0;
		} else {
			goto failure;
		}
	}

	if(!server) {
		debug("mover: opening connection to %s\n",hop);
		server = kangaroo_server_open(hop,0);
		if(!server) goto failure;
		strcpy(last_hop,hop);
	}

	result = kangaroo_message_send(server,m);
	if(!result) goto failure;

	success:
	messages_out++;
	kangaroo_history_outgoing_add(m->length);
	kangaroo_message_delete(m);
	return 1;

	failure:
	save_errno = errno;
	debug("mover: couldn't send to %s: %s\n",hop,strerror(errno));
	debug_wait();
	if(m) {	
		kangaroo_message_delete(m);
		m=0;
	}	
	errno = save_errno;
	return 0;
}

void mover_thread()
{
	int delay;
	int do_commit = 0;
	int save_errno;
	struct kangaroo_message *m;

	last_commit = time(0);
	delay = kangaroo_config_get_int("kangaroo.mover.timeout.min");

	while(1) {
		do_commit = 0;
		m = 0;

		if(pause_mode) {
			debug("mover: paused at user request.\n");
			while(pause_mode) {
				sleep(1);
			}
			debug("mover: resumed at user request.\n");
		}

		if( messages_out ) {
			if( !message_list_consume_peek(kangaroo_message_list) ) {
				debug("mover: empty list, so committing\n");
				do_commit = 1;
			} else if( (time(0)-last_commit) > kangaroo_config_get_int("kangaroo.mover.timeout.periodic")) {
				debug("mover: periodic commit\n");
				do_commit = 1;
			}
		}

		if(pause_mode) {
			errno = EINTR;
			goto failure;
		}

		if(do_commit) {
			if(server && kangaroo_commit(server)) {
				goto failure;
			} else {
				mover_commit();
			}
		}

		m = message_list_consume(kangaroo_message_list);

		if( m->type==KANGAROO_RENAME ) {
			debug("mover: forcing commit before rename\n");
			if(server && kangaroo_commit(server)) {
				goto failure;
			} else {
				mover_commit_without_delete();
			}	
		}

		if(mover_send(m)) {
			delay = kangaroo_config_get_int("kangaroo.mover.timeout.min");

			if( m->type==KANGAROO_RENAME ) {
				debug("mover: forcing commit after rename\n");
				if(server && kangaroo_commit(server)) {
					goto failure;
				} else {
					mover_commit();
				}
			}
		} else {
			m = 0;
			goto failure;
		}

		continue;

		failure:
		save_errno = errno;
		
		if(m) {
			error_state_set(m->addr,m->path,"couldn't send '%s %s' to %s because %s",kangaroo_message_name(m),m->path,m->addr,strerror(save_errno));
		} else {
			char host[KANGAROO_HOST_MAX];
			if((server) && (!kangaroo_server_address(server,host))) {
				strcpy(host,"unknown");
			}
			error_state_set(host,"","couldn't send 'commit' to %s because %s",host,strerror(save_errno));
		}
		debug("mover: try again in %d seconds\n",delay);
		mover_abort();
		timestamp_sleep(delay*1000000);
		delay = MIN(delay*2,kangaroo_config_get_int("kangaroo.mover.timeout.max"));
	}
}

