
#include "network.h"
#include "network_cache.h"
#include "hash_table.h"
#include "debug.h"

#include <pthread.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>

#define CACHE_SIZE 127
#define CACHE_LIFETIME 3600

static struct hash_table *name_to_addr=0;
static struct hash_table *addr_to_name=0;

static pthread_mutex_t name_to_addr_mutex = PTHREAD_MUTEX_INITIALIZER;
static pthread_mutex_t addr_to_name_mutex = PTHREAD_MUTEX_INITIALIZER;

static int cache_init()
{
	pthread_mutex_lock(&name_to_addr_mutex);
	if(!name_to_addr) {
		name_to_addr = hash_table_create(127,hash_string);	
		if(!name_to_addr) {
			pthread_mutex_unlock(&name_to_addr_mutex);
			return 0;
		}
	}
	pthread_mutex_unlock(&name_to_addr_mutex);

	pthread_mutex_lock(&addr_to_name_mutex);	
	if(!addr_to_name) {
		addr_to_name = hash_table_create(127,hash_string);	
		if(!addr_to_name) {
			pthread_mutex_unlock(&addr_to_name_mutex);
			return 0;
		}
	}
	pthread_mutex_unlock(&addr_to_name_mutex);

	return 1;
}

int network_cache_name_to_addr( const char *name, char *addr )
{
	char *found, *copy;
	network_address a;
	int success;

	if(!cache_init()) return 0;

	pthread_mutex_lock(&name_to_addr_mutex);	

	found = hash_table_lookup(name_to_addr,name);
	if(found) {
		strcpy(addr,found);
		pthread_mutex_unlock(&name_to_addr_mutex);
		return 1;
	}

	pthread_mutex_unlock(&name_to_addr_mutex);

	debug("network_cache: looking up name %s...\n",name);

	success = network_name_to_address(name,&a);
	if(success) {
		network_address_to_string(a,addr);
		debug("network_cache: %s is %s\n",name,addr);
	} else {
		debug("network_cache: couldn't lookup %s: %s\n",name,strerror(errno));
		return 0;
	}

	copy = strdup(addr);
	if(!copy) return 1;

	pthread_mutex_lock(&name_to_addr_mutex);
	success = hash_table_insert(name_to_addr,name,copy,&found);
	if(found) free(found);
	pthread_mutex_unlock(&name_to_addr_mutex);

	return 1;
}

int network_cache_addr_to_name( const char *addr, char *name )
{
	char *found, *copy;
	network_address a;
	int success;

	if(!cache_init()) return 0;

	pthread_mutex_lock(&addr_to_name_mutex);	

	found = hash_table_lookup(addr_to_name,addr);
	if(found) {
		strcpy(name,found);
		pthread_mutex_unlock(&addr_to_name_mutex);
		return 1;
	}

	pthread_mutex_unlock(&addr_to_name_mutex);

	debug("network_cache: looking up address %s...\n",addr);

	network_string_to_address(addr,&a);
	success = network_address_to_name(a,name);
	if(success) {
		debug("network_cache: %s is %s\n",addr,name);
	} else {
		debug("network_cache: couldn't lookup %s: %s\n",addr,strerror(errno));
		return 0;
	}

	copy = strdup(name);
	if(!copy) return 1;

	pthread_mutex_lock(&addr_to_name_mutex);
	success = hash_table_insert(addr_to_name,addr,copy,&found);
	if(found) free(found);
	pthread_mutex_unlock(&addr_to_name_mutex);

	return 1;
}
