/*
Kangaroo is Copyright (C) 2001 Douglas Thain.
Permission is granted to use the source code for research and study purposes.
Other uses must be cleared with the copyright holder.
*/


#include "routing_table.h"
#include "network.h"
#include "stringtools.h"
#include "debug.h"
#include "network_cache.h"

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>

#define ROUTING_TABLE_LINE_MAX 256

struct routing_table {
	char *source;
	char *dest;
	char *hop;
	struct routing_table *next;
};

struct routing_table_iterator {
  struct routing_table *first;
  struct routing_table *next;
};

static struct routing_table * routing_table_create( char *s, char *d, char *h, struct routing_table *n )
{
	struct routing_table *r;

	r = malloc(sizeof(*r));
	if(!r) return 0;

	r->source = strdup(s);
	if(!r->source) {
		free(r);
		return 0;
	}

	r->dest = strdup(d);
	if(!r->dest) {
		free(r->source);
		free(r);
		return 0;
	}

	r->hop = strdup(h);
	if(!r->hop) {
		free(r->source);
		free(r->dest);
		free(r);
		return 0;
	}

	r->next = n;

	return r;
}

void routing_table_delete( struct routing_table *r )
{
	if(r->next) routing_table_delete(r->next);
	free(r->source);
	free(r->dest);
	free(r->hop);
	free(r);
}

static int canonicalize( char *name )
{
	char addr_string[NETWORK_ADDR_MAX];

	/* If it's a special rule, leave it along */
	if(name[0]=='@') return 1;

	/* If it's a wildcard, leave it alone */
	if(strchr(name,'*')) return 1;

	/* Otherwise, convert it into an address string */	
	if(!network_cache_name_to_addr(name,addr_string)) {
		return 0;
	} else {
		strcpy(name,addr_string);
		return 1;
	}
}

struct routing_table * routing_table_load( const char *filename )
{
	struct routing_table *r=0,*s=0;
	char source[ROUTING_TABLE_LINE_MAX];
	char dest[ROUTING_TABLE_LINE_MAX];
	char hop[ROUTING_TABLE_LINE_MAX];
	char line[ROUTING_TABLE_LINE_MAX];
	int fields;
	int linenum=0;
	int save_errno;
	FILE *f=0;

	network_address addr;
	char my_name[NETWORK_NAME_MAX];
	char my_addr[NETWORK_ADDR_MAX];

	if(!network_name_get(my_name)) return 0;
	if(!network_address_get(&addr)) return 0;
	network_address_to_string(addr,my_addr);

	f = fopen(filename,"r");
	if(!f) goto failure;

	while(fgets(line,ROUTING_TABLE_LINE_MAX,f)) {

		linenum++;

		if(line[0]=='#') continue;
		line[strlen(line)-1]=0;

		fields = sscanf(line,"%s %s %s",source,dest,hop);
		if(fields<=0) {
			continue;
		} else if(fields==3) {
			/* Skip rules that don't apply to me */
			if( !pattern_match(source,my_addr) && !pattern_match(source,my_name) ) continue;
			if(!canonicalize(hop)) goto failure;
			s = routing_table_create(source,dest,hop,r);
			if(!s) goto failure;
			r = s;
		} else {
			goto parse_error;
		}
	}

	fclose(f);
	return r;

	parse_error:
	fprintf(stderr,"%s: parse error on line %d\n",filename,linenum);

	failure:
	save_errno = errno;
	if(f) fclose(f);
	if(r) routing_table_delete(r);
	errno = save_errno;
	return 0;
}

/*
Search a given routing table for a rule governing 'dest'.
On success, return 1 and fill 'hop' with the next hop.
On failure, return 0 with errno set.
Special values for 'hop' may be:
    STORE   Deliver data to this filesystem.
    HOLD    Hold data in this spool.  (for now.)
*/

int routing_table_get_hop( struct routing_table *r, const char *dest, char *hop )
{
	char source_name[NETWORK_NAME_MAX];
	char source_addr[NETWORK_ADDR_MAX];
	char dest_name[NETWORK_NAME_MAX];
	char dest_addr[NETWORK_ADDR_MAX];
	network_address addr;

	/* First look up my address and name.  This is cached, so it's cheap. */

	if(!network_name_get(source_name)) return 0;
	if(!network_address_get(&addr)) return 0;
      	network_address_to_string(addr,source_addr);

	/* Now, canonicalize the destination.  Also cached.  */

	if(!network_cache_name_to_addr(dest,dest_addr)) return 0;
	if(!network_cache_addr_to_name(dest_addr,dest_name)) return 0;

	/* Implicit rule: if source==dest, then hop="STORE" */

	if(!strcmp(source_addr,dest) || !strcmp(dest_addr,"127.0.0.1") || !strcmp(dest_addr,"0.0.0.0")) {
		strcpy(hop,"@STORE");
		return 1;
	}

	/* Now consider all of the rules.  The source address will be IP addresses or domain names, possibly wildcarded.  The destinations will be IP addresses, possibly wildcarded.  We must compare all fields in all forms.  */

	while(r) {
		if(
			(
				pattern_match(r->source,source_name) ||
				pattern_match(r->source,source_addr)
			) && (
				pattern_match(r->dest,dest_name) ||
				pattern_match(r->dest,dest_addr)
			)
		) {
			if (strcmp(r->hop, "@DIRECT") == 0) {
				strcpy(hop,dest);
			} else {
				strcpy(hop,r->hop);
			}
				return 1;
		}
		r = r->next;
	}

	errno = EHOSTUNREACH;
	return 0;
}

/*
Function that returns multiple routes to a destination
*/

struct routing_table* routing_table_get_all_hops (struct routing_table *r, const char *dest)
{
	char source_name[NETWORK_NAME_MAX];
	char source_addr[NETWORK_ADDR_MAX];
  char dest_name[NETWORK_NAME_MAX];
  char dest_addr[NETWORK_ADDR_MAX];
	network_address addr;
	struct routing_table *ptable;
	struct routing_table *plist;
	int found;
	
	/* First look up my address and name.  This is cached, so it's cheap. */
	if(!network_name_get(source_name)) return 0;
	if(!network_address_get(&addr)) return 0;
      	network_address_to_string(addr,source_addr);	
      	
  /* Now, canonicalize the destination.  Also cached.  */
  if(!network_cache_name_to_addr(dest,dest_addr)) return 0;
  if(!network_cache_addr_to_name(dest_addr,dest_name)) return 0;
	
	/* Implicit rule: if source==dest, then hop="STORE" */
	if(!strcmp(source_addr,dest) || !strcmp(dest,"127.0.0.1") || !strcmp(dest,"0.0.0.0")) {
		/* This is the destination, so only one node in the return value */
		ptable = malloc(sizeof(struct routing_table));
		ptable->source = strdup(dest);
		ptable->dest   = strdup(dest);
		ptable->hop    = strdup("@STORE");
		ptable->next = 0;
		plist = ptable;
		return plist;
	}
	
	ptable = 0;
  plist  = 0;
	
	/* Now consider all of the rules.  The source address will be IP addresses or domain names, possibly wildcarded.
	The destinations will be IP addresses, possibly wildcarded.  We must compare all fields in all forms.  */
	found = 0;
	while(r) {
		if(
			(
				pattern_match(r->source,source_name) ||
				pattern_match(r->source,source_addr)
			) && (
				pattern_match(r->dest,dest_name)	   ||
				pattern_match(r->dest,dest_addr)
			)
		) {
				if (strcmp(r->hop, "@DIRECT") == 0)  {
					if (found != 1)	{/* If no other routes exist, then only use DIRECT */
						ptable = malloc(sizeof(struct routing_table));
						ptable->hop  = strdup(dest);
						ptable->source = strdup(source_name);
						ptable->dest   = strdup(dest);
						ptable->next = plist;
						plist = ptable;
						//debug ("Inserting hop %s for dest %s found is %d \n", ptable->hop, ptable->dest,found);
					}	
				} else {
					found = 1;
					ptable = malloc(sizeof(struct routing_table));
					ptable->hop  = strdup(r->hop);
					ptable->source = strdup(source_name);
					ptable->dest   = strdup(dest);
					ptable->next = plist;
					plist = ptable;
					//debug ("Inserting hop %s for dest %s next is %p\n", ptable->hop, ptable->dest, ptable->next);
				}//end else
			}//pattern match
		r = r->next;
	}

	if (plist == 0)	{
		errno = EHOSTUNREACH;
	}	
	return plist;
	
}

/*This function cleans up all entries in the list */
void routing_table_cleanup (struct routing_table *plist)	
{
	struct routing_table *tmp = plist;
	
	while (tmp)	{
		plist = tmp;
		free (plist->source);
		free (plist->dest);
		free (plist->hop);
		tmp = tmp->next;
		free (plist);
	}
	plist = 0;
}

/* Function to initialize iterator */
struct routing_table_iterator* routing_table_iterator_init (struct routing_table *plist)
{
	struct routing_table_iterator* piterator;
	piterator = malloc(sizeof(struct routing_table_iterator));
	piterator->next = piterator->first = plist;
	return piterator;
}

void routing_table_iterator_destroy (struct routing_table_iterator *piterator)
{
	free (piterator);
}

/* Function iterates through routing table entry */
char* routing_table_iterator_nexthop (struct routing_table_iterator *piterator)
{
	char *pnxthop = 0;
	
	if ((piterator) && (piterator->next))	{
	 pnxthop   = piterator->next->hop;
	 piterator->next = piterator->next->next;
	}
	//debug ("Next hop is:%s\n", pnxthop);
	return pnxthop;
}

int routing_table_num_routes (const struct routing_table *plist)	
{
	int count;
	const struct routing_table *tmp = plist;
	
	if (plist == 0)	{
		return -1;
	}
	
	count = 0;
	while (tmp)	{
		if (!strcmp (tmp->hop, "@STORE"))	{
			return 1;		//This is the destination!!
		}
		count++;		
		//debug ("Hop %d is %s\n",count, tmp->hop);
		tmp = tmp->next;
	}
	
	return count;	
}
