/*
Kangaroo is Copyright (C) 2001 Douglas Thain.
Permission is granted to use the source code for research and study purposes.
Other uses must be cleared with the copyright holder.
*/


#include "spool_dir.h"
#include "kangaroo_types.h"
#include "kangaroo_config.h"
#include "reaper.h"
#include "debug.h"

#include <string.h>
#include <stdio.h>
#include <time.h>
#include <string.h>
#include <stdlib.h>

#include <unistd.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>

static char *spool_dir=0;
int spool_dir_space_remaining = 0;

int spool_dir_init( const char *dir )
{
	spool_dir_space_remaining = atoi(kangaroo_config_get( "kangaroo.spool.max") );	
	spool_dir = strdup(dir);
	if(!spool_dir) return 0;

	if( mkdir(dir,0700)<0 && errno!=EEXIST ) {
		return 0;
	}
  debug ("Spool dir space initialized to: %d\n", spool_dir_space_remaining);
	return 1;
}

static FILE * spool_dir_create( char *filename )
{
	struct timeval tv;
	FILE *file;
	int fd;
	int i;

	for(i=0;i<100;i++) {
		if(gettimeofday(&tv,0)) return 0;

		sprintf(filename,"%s/%ld-%ld",spool_dir,tv.tv_sec,tv.tv_usec);

		fd = open(filename,O_CREAT|O_EXCL|O_WRONLY,0700);
		if(fd<0) continue;

		file = fdopen(fd,"wb");
		if(!file) {
			close(fd);
			reaper_queue(filename);
			continue;
		}

		return file;
	}

	return 0;
}

int spool_dir_insert( char *tagout, struct kangaroo_message *m )
{
	char filename[KANGAROO_PATH_MAX];
	char commitname[KANGAROO_PATH_MAX];
	FILE *file=0;
	int space_reqd;
	int result;
	char *tag;
	int save_errno;

	filename[0] = 0;
	commitname[0] = 0;

	file = spool_dir_create(filename);
	if(!file) goto failure;

	tag = strrchr(filename,'/')+1;
	
	space_reqd = kangaroo_space_required(m,1);
  if (space_reqd > spool_dir_space_remaining) {
    debug ("No space on spool device: Reqd %d Space left %d \n", space_reqd, spool_dir_space_remaining);
    errno = ENOSPC;
	fclose(file);
	reaper_queue (filename);
    return 0;
  }

  debug ("Space required for write is %d %d\n", space_reqd, m->length);
	result = kangaroo_message_write(file,m,1);
	if(!result) goto failure;

 	spool_dir_space_remaining -= result;
 	debug ("spool dir space remaining:%d \n", spool_dir_space_remaining);
	
	result = fclose(file);
	if(result!=0) goto failure;

	file = 0;

	sprintf(commitname,"%s.commit",filename);
	result = rename(filename,commitname);
	if(result!=0) goto failure;

	strcpy(tagout,tag);
	return 1;

	failure:
	save_errno = errno;
	if(file) fclose(file);
	reaper_queue(commitname);
	reaper_queue(filename);
	errno = save_errno;
	debug ("Problem in spool dir insert because %s\n", strerror(errno));
	return 0;
}

struct kangaroo_message * spool_dir_lookup( const char *tag, int want_data )
{
	struct kangaroo_message *m;
	char filename[KANGAROO_PATH_MAX];
	FILE *file=0;

	memset (filename, 0, sizeof(filename));
	sprintf(filename,"%s/%s.commit",spool_dir,tag);

	file = fopen(filename,"rb");
	if(!file) goto failure;

	m = kangaroo_message_read( file, want_data );
	if(!m) goto failure;

	strcpy(m->tag,tag);

	fclose(file);

	return m;

	failure:
	if(file) fclose(file);
	return 0;
}

int spool_dir_remove( const char *tag )
{
	char filename[KANGAROO_PATH_MAX];
	int fret;
	struct stat buf;
	
	sprintf(filename,"%s/%s.commit",spool_dir,tag);
	fret = stat (filename, &buf);
	if (fret != -1)	{
		spool_dir_space_remaining += buf.st_size;
		debug ("spool dir space remaining after increase:%d\n", spool_dir_space_remaining);
	}
	else	{
		 debug ("stat returned -1, because %s\n", strerror(errno));
	}
	reaper_queue(filename);

	return 1;
}

static int spool_dir_compare( const void *va, const void *vb )
{
	time_t ta,tb;
	long ua,ub;

	int fields;
	char **a, **b;

	a = (char **)va;
	b = (char **)vb;

	fields = sscanf(*a,"%ld-%ld",&ta,&ua);
	if(fields!=2) return 0;

	fields = sscanf(*b,"%ld-%ld",&tb,&ub);
	if(fields!=2) return 0;

	if(ta<tb) {
		return -1;
	} else if(ta==tb) {
		if(ta<tb) {
			return -1;
		} else if(ua==ub) {
			return 0;
		} else {
			return 1;
		}
	} else {
		return 1;
	}
}

int spool_dir_iterate( int (*iterator) ( char *filename, char *tag, void *arg ), void *arg )
{
	int i;
	int items=0;
	int max=100;
	char **list;
	char **newlist;
	char filename[KANGAROO_PATH_MAX];
	DIR *dir;
	struct dirent *d;
	char *dot;
	int keepgoing=1;

	list = malloc( sizeof(char*)*max );
	if(!list) return -1;

	dir = opendir(spool_dir);
	if(!dir) return -1;

	while( (d=readdir(dir)) ) {
		if(d->d_name[0]=='.') continue;

		list[items] = strdup(d->d_name);
		if(!list[items]) break;

		items++;

		if(items>=max) {
			newlist = malloc(sizeof(char*)*max*2);
			if(!newlist) break;

			for(i=0;i<max;i++) {
				newlist[i] = list[i];
			}

			for(i=max;i<(max*2);i++) {
				newlist[i] = 0;
			}

			free(list);
			list = newlist;
			max = max*2;
		}
	}
	closedir(dir);

	if(items>0) qsort(list,items,sizeof(*list),spool_dir_compare);

	for( i=0; i<items; i++ ) {
		sprintf(filename,"%s/%s",spool_dir,list[i]);
		dot = strrchr(list[i],'.');
		if(dot) {
			*dot = 0;
			if(keepgoing) {
				keepgoing = iterator( filename, list[i], arg );
			}
		}
		free(list[i]);
	}
	free(list);

	return keepgoing;
}
